"""
*********************************************************************************
*                                                                               *
* web_parsing_model.py -- A web parsing job info model                          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.db import models

from black_widow.app.gui.web.black_widow.models.abstract_job_model import AbstractJobModel
from black_widow.app.managers.parser import HtmlParser


class WebParsingJobModel(AbstractJobModel):
    """
    Django Web Parsing Job Model
    """
    TYPE_SINGLE_PAGE = 'single_page'
    TYPE_WEBSITE_CRAWLING = 'website_crawling'
    TYPES = (
        (TYPE_SINGLE_PAGE, 'Single Page'),
        (TYPE_WEBSITE_CRAWLING, 'Website Crawling')
    )

    PARSE_TAGS = (
        (HtmlParser.TYPE_ALL, 'All Tags'),
        (HtmlParser.TYPE_RELEVANT, 'Relevant Tags (a, script, link, Form Tags, Meta Tags)'),
        (HtmlParser.TYPE_FORM, 'Form Tags (form, input, textarea, select, option)'),
        (HtmlParser.TYPE_META, 'Meta Tags (meta, xmp)')
    )

    url: str = models.CharField(null=False, max_length=512)
    _parsing_type: str = models.CharField(
        null=False,
        choices=TYPES,
        max_length=50
    )
    _parsing_tags: str = models.CharField(
        null=False,
        choices=PARSE_TAGS,
        max_length=50
    )
    depth: int = models.IntegerField(null=False)
    cookies: str = models.TextField(null=True)

    @staticmethod
    def all() -> models.query.QuerySet:
        return AbstractJobModel._all(WebParsingJobModel)

    def parsing_type_key(self):
        return self._parsing_type

    @property
    def parsing_type(self):
        return dict(WebParsingJobModel.TYPES).get(self._parsing_type)

    @parsing_type.setter
    def parsing_type(self, value: str):
        self._parsing_type = value

    def parsing_tags_key(self):
        return self._parsing_tags

    @property
    def parsing_tags(self):
        return dict(WebParsingJobModel.PARSE_TAGS).get(self._parsing_tags)

    @parsing_tags.setter
    def parsing_tags(self, value: str):
        self._parsing_tags = value

    def __str__(self) -> str:
        return 'WebParsingJobModel(' + str({
            'id': self.id,
            'json_file': self.json_file,
            'pid': self.pid,
            'pid_file': self.pid_file,
            'status': self.status_name,
            'type': self.parsing_type,
            'tags': self.parsing_tags,
            'depth': self.depth,
        }) + ')'



"""
*********************************************************************************
*                                                                               *
* pcap_layer_field.py -- Packet Capture Layer Field                             *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from anytree import Node
from pyshark.packet.fields import LayerField


class PcapLayerField(Node):
    """
    Packet Capture Layer Field
    """

    SRC_FIELDS = ('src', 'src_host', 'src_hw_mac', 'src_proto_ipv4', 'ta', 'sa', 'srcport')
    DST_FIELDS = ('dst', 'dst_host', 'dst_hw_mac', 'dst_proto_ipv4', 'ra', 'da', 'dstport')

    PROTO_FIELDS = ('proto', 'phy', 'request_version', 'response_version')

    AMBIGUOUS_FIELD_NAMES = (
        'ip.addr', 'ip.host',
        'ipv6.addr', 'ipv6.host',
        'udp.port',
        'tcp.port',
        'tcp.segment_data'
    )

    def __init__(self, layer_field: LayerField = None, sanitized_name: str = None, parent=None, children=None):
        """
        :param layer_field: The LayerField of this Node
        :type parent: PcapLayerField
        :param children: The children of this Node
        """
        kwargs = {}
        if layer_field is not None:
            kwargs.update({
                'field': layer_field
            })
            name = layer_field.name
            self.is_main = False
        else:
            name = 'root'
            self.is_main = True
        self.field = layer_field
        self.sanitized_name = sanitized_name
        super().__init__(name, parent, children, **kwargs)

    @property
    def label(self):
        if self.field is None:
            return self.name
        field_label: str = self.field.showname_key
        if not field_label:
            return self.name
        return field_label

    @property
    def value(self):
        if self.field is None:
            return self.name
        field_value: str = self.field.showname_value
        if not field_value:
            field_value = self.field.get_default_value()
        if field_value is not None:
            field_value = field_value.encode('utf-8').decode('unicode-escape').replace('Âµs', 'µs')
        return field_value

    @property
    def size(self) -> float or None:
        if self.field is None:
            return 0
        return float(self.field.size)

    @property
    def pos(self) -> int or None:
        if self.field is None:
            return 0
        return int(self.field.pos)

    def get_dict(self) -> dict:
        node_dict = {
            'children': []
        }
        if not self.is_main:
            node_dict.update({
                'label': self.label,
                'value': self.value,
                'pos': self.pos,
                'key': self.name,
                'size': self.size,   # byte
                'name': self.sanitized_name
            })
        for child in self.children:
            child: PcapLayerField
            node_dict['children'].append(child.get_dict())
        return node_dict

    def __str__(self, depth: int = 0):
        """
        :param depth: The current printing depth
        """
        pcap_layer_field_header = '   |' * depth
        pcap_layer_field_row = pcap_layer_field_header + '   ├── [ ' + str(self.label) + ' ]'
        if self.value is not None:
            value = str(self.value)
            if len(value) > 30:
                value = value[0:30] + '...'
            pcap_layer_field_row += ' = ' + value
        else:
            pcap_layer_field_row += ' : PARENT NAME'
        pcap_layer_field_row += ', key=' + self.name + ', name' + self.sanitized_name + \
                                ', size=' + str(self.size) + ', pos=' + str(self.pos)
        for pcap_layer_field_child in self.children:
            pcap_layer_field_child: PcapLayerField
            pcap_layer_field_row += "    \n" + pcap_layer_field_child.__str__(depth+1)
        if len(self.children) > 0:
            pcap_layer_field_row += "   |\n" + pcap_layer_field_header
        return pcap_layer_field_row



"""
*********************************************************************************
*                                                                               *
* settings.py -- Django settings for web project.                               *
*                                                                               *
* Generated by 'django-admin startproject' using Django 1.11.22.                *
*                                                                               *
* For more information on this file, see                                        *
* https://docs.djangoproject.com/en/1.11/topics/settings/                       *
*                                                                               *
* For the full list of settings and their values, see                           *
* https://docs.djangoproject.com/en/1.11/ref/settings/                          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os

from black_widow.app.env import APP_STORAGE, APP_WEB, APP_DEBUG
from black_widow.app.helpers.network import get_ip_address
from black_widow.app.gui.web.wsgi import WEB_PACKAGE

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = APP_STORAGE

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.11/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = '2-2+=!(i*_ute^&^j=7*t&mgaamgod+=l*8c#w84jxm@!qgp^g'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = APP_DEBUG

ALLOWED_HOSTS = ['0.0.0.0', '127.0.0.1', 'localhost', get_ip_address()]

# Application definition

INSTALLED_APPS = [
    # 'django.contrib.admin',
    # 'django.contrib.auth',
    'django.contrib.contenttypes',
    # 'django.contrib.sessions',
    # 'django.contrib.messages',
    'django.contrib.staticfiles',
    'black_widow.app.gui.web.black_widow'
]

MIDDLEWARE = [
    # 'django.middleware.security.SecurityMiddleware',
    # 'django.contrib.sessions.middleware.SessionMiddleware',
    'django.middleware.common.CommonMiddleware',
    # 'django.middleware.csrf.CsrfViewMiddleware',
    # 'django.contrib.auth.middleware.AuthenticationMiddleware',
    # 'django.contrib.messages.middleware.MessageMiddleware',
    'django.middleware.clickjacking.XFrameOptionsMiddleware',
]

ROOT_URLCONF = WEB_PACKAGE + '.urls'

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [
            os.path.join(APP_WEB, 'templates')
        ],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                #         'django.template.context_processors.debug',
                'django.template.context_processors.request'
                #         'django.contrib.auth.context_processors.auth',
                #         'django.contrib.messages.context_processors.messages',
            ],
        },
    },
]

WSGI_APPLICATION = WEB_PACKAGE + '.wsgi.application'

# Database
# https://docs.djangoproject.com/en/1.11/ref/settings/#databases

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': os.path.join(BASE_DIR, 'black-widow.sqlite3'),
    }
}

# Password validation
# https://docs.djangoproject.com/en/1.11/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        'NAME': 'django.contrib.auth.password_validation.UserAttributeSimilarityValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.MinimumLengthValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.CommonPasswordValidator',
    },
    {
        'NAME': 'django.contrib.auth.password_validation.NumericPasswordValidator',
    },
]

# Internationalization
# https://docs.djangoproject.com/en/1.11/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.11/howto/static-files/

STATIC_URL = '/static/'
STATICFILES_DIRS = [
    os.path.join(APP_WEB, 'static')
]



"""
*********************************************************************************
*                                                                               *
* sql_view.py -- Django SQL Injection views.                                    *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.shortcuts import render, redirect

from black_widow.app.services import Log

from .abstract_view import AbstractView


class Sql:
    """
    SQL injection Container View
    """
    class SettingsView(AbstractView):
        """
        SQL View
        """
        name = 'sql-injection'
        # template_name = 'sql/settings.html'
        template_name = 'error/not_implemented.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """
            # TODO: get current sql injection jobs
            view_params = dict()
            return render(request, self.template_name, view_params)

        def post(self, request):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponseRedirect
            """
            # TODO: create new SQL injection job
            job_id = 0
            return redirect('sql/inject?job_id=' + str(job_id))

    class InjectView(AbstractView):
        """
        Injection View
        """
        name = 'sniffing'
        # template_name = 'sql/inject.html'
        template_name = 'error/not_implemented.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """
            request_params: dict = request.GET.dict()
            job_id = request_params.get('job_id')
            Log.info("Showing job #" + str(job_id))
            return render(request, self.template_name)



"""
*********************************************************************************
*                                                                               *
* html_parser.py -- HTML/URL parsing and Website crawling manager               *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from abc import ABC
from html.parser import HTMLParser as PyHTMLParser
from urllib.parse import urlparse
from tidylib import tidy_document

from black_widow.app.managers.request import HttpRequest
from black_widow.app.services import Log, JsonSerializer
from black_widow.app.helpers.validators import is_url
from black_widow.app.helpers.util import is_listable


class HtmlParser(PyHTMLParser, ABC):
    """
    HtmlParser Manager
    """

    TYPE_ALL = 'all_parse'
    TYPE_RELEVANT = 'relevant_parse'
    TYPE_FORM = 'form_parse'
    TYPE_META = 'meta_parse'

    _input_tags = {
        'input': [
            'id', 'name', 'type',
            'min', 'max',
            'required',
            'minlength',
            'maxlength',
            'pattern',
            'value'
        ],
        'textarea': [
            'id', 'name', 'data',
            'required',
            'minlength',
            'maxlength'
        ],
        'select': [
            'id', 'name', 'required'
        ],
        'option': [
            'id', 'value', 'selected'
        ],
        'button': [
            'id', 'name', 'type', 'form', 'value'
        ]
    }

    _xmp_tags = {
        'x:xmpmeta': ['xmlns:x', 'x:xmptk'],
        'rdf:rdf': ['xmlns:rdf'],
        'rdf:description': [
            'data',
            'rdf:about',
            'xmlns:dc',
            'xmlns:xmp',
            'xmlns:xmpRights',
            'xmp:CreatorTool',
            'xmpRights:Marked',
            'xmpRights:Marked'
        ],
        'rdf:seq': ['data'],
        'rdf:li': ['data'],
        'rdf:alt': ['data'],
        'dc:contributor': ['data'],
        'dc:coverage': ['data'],
        'dc:creator': ['data'],
        'dc:date': ['data'],
        'dc:description': ['data'],
        'dc:format': ['data'],
        'dc:identifier': ['data'],
        'dc:language': ['data'],
        'dc:publisher': ['data'],
        'dc:relation': ['data'],
        'dc:rights': ['data'],
        'dc:source': ['data'],
        'dc:subject': ['data'],
        'dc:title': ['data'],
        'dc:type': ['data']
    }

    _meta_tags = {
        'meta': ['name', 'itemprop', 'property', 'content', 'charset']
    }
    _meta_tags.update(_xmp_tags)

    _tag_names = ('name', 'itemprop', 'property', 'id')
    _tag_key_names = ('content', 'charset')

    _relevant_tags = {
        'a': ['href'],
        'img': ['src'],
        'form': ['id', 'action', 'method', 'name'],
        'script': ['src', 'data', 'type'],  # { 'src': '/some/script.js', 'data': 'function() ...' }
        'link': ['href'],  # { 'href': '*.xml' }
        'html': [],
        'body': []
    }
    _relevant_tags.update(_input_tags)
    _relevant_tags.update(_meta_tags)

    # Attributes witch contain URLs
    _url_attrs = ['href', 'src', 'action']

    # Not closed tags
    _not_closed_tags = ['input', 'link', 'meta', 'hr', 'img', 'br', 'source']

    _unacceptable_content_types = ['application/zip']

    def __init__(self, relevant: bool = False):
        super().__init__()
        self.tags = {}
        self.relevant = relevant
        self.queue_tag_ignored = []
        self.queue_tag = []
        self.queue_form = []
        self.url = None
        self.base_url = None
        self.url_scheme = ''

    @staticmethod
    def crawl(url: str, parsing_type: str, callback, depth: int = 0, cookies: str = None):
        """
        :param url: The url to crawl/parse
        :param parsing_type: HtmlParse.TYPE_ALL | HtmlParse.TYPE_RELEVANT | HtmlParse.TYPE_FORM | HtmlParse.TYPE_META
        :param callback: The callback method to call foreach visited page
        :param depth: The max crawling depth (0 to execute a normal page parsing, < 0 for no limit)
        :param cookies: The cookies to use on parsing
        """
        if not is_url(url):
            raise ValueError('url must be a valid url')
        if parsing_type not in HtmlParser.types():
            raise ValueError('parsing_type must be one of ' + str(HtmlParser.types()))
        if not callable(callback):
            raise ValueError('callback is not callable')
        if type(depth) is not int:
            raise ValueError('dept must be an integer')
        if cookies is not None and type(cookies) is not str:
            raise ValueError('cookies must be a string')

        base_url = urlparse(url).netloc
        base_urls = (base_url, )
        if base_url[0:4] != 'www.':
            base_urls += ('www.' + str(base_url),)
        parsed_urls = set()
        parsed_hashes = set()

        def _crawl(href: str, curr_depth: int = 0):
            parsed_href = urlparse(href)
            href = parsed_href.scheme + '://' + parsed_href.netloc + parsed_href.path
            if parsed_href.query != '':
                href += '?' + parsed_href.query
            if href in parsed_urls or \
                    parsed_href.netloc not in base_urls or \
                    'mailto:' in parsed_href.path or \
                    (0 <= depth and (depth < curr_depth)):
                return

            # Visit the current href
            if parsing_type == HtmlParser.TYPE_ALL:
                parsed, _ = HtmlParser.all_parse(href, cookies=cookies)
            else:
                parsed, _ = HtmlParser.relevant_parse(href, cookies=cookies)

            parsed_hash = hash(JsonSerializer.dumps_json(parsed))
            if parsed_hash in parsed_hashes:
                return

            parsed_hashes.add(parsed_hash)
            parsed_urls.add(href)

            if parsing_type == HtmlParser.TYPE_FORM:
                # Find forms in page
                parsed_page = HtmlParser.find_forms(parsed, href)
            elif parsing_type == HtmlParser.TYPE_META:
                # Find metadata in page
                parsed_page = HtmlParser.find_meta(parsed)
            else:
                parsed_page = parsed

            if parsed_page.get('tag') is not None:
                parsed_page = {
                    0: parsed_page
                }

            parsed_page['url'] = href
            callback(parsed_page)

            # Find adjacent links
            links = HtmlParser.find_links(parsed)
            for link in links:
                _crawl(link, curr_depth + 1)

        _crawl(url)
        Log.success(url + ' crawling done!')

    @staticmethod
    def all_parse(url: str = None, html: str = None, cookies: str = None) -> (dict, str):
        """
        Make an HTML/URL parsing by processing ALL found tags
        :param url: The url to parse (or None)
        :param html: The html page to parse as string (or None)
        :param cookies: The cookies to use on parsing
        :return: dictionary of tags, cookies
        """
        return HtmlParser.__abstract_parse(url, html, False, cookies)

    @staticmethod
    def relevant_parse(url: str = None, html: str = None, cookies: str = None) -> (dict, str):
        """
        Make an HTML/URL parsing by processing only tags in HtmlParse._relevant_tags
        :param url: The url to parse (or None)
        :param html: The html page to parse as string (or None)
        :param cookies: The cookies to use on parsing
        :return: dictionary of tags, cookies
        """
        return HtmlParser.__abstract_parse(url, html, True, cookies)

    @staticmethod
    def form_parse(url: str = None, html: str = None, cookies: str = None) -> (dict, str):
        """
        Make an HTML/URL parsing by processing only tags in HtmlParse._relevant_tags
        :param url: The url to parse (or None)
        :param html: The html page to parse as string (or None)
        :param cookies: The cookies to use on parsing
        :return: dictionary of form tags, cookies
        """
        parsed_html, cookies = HtmlParser.relevant_parse(url, html, cookies)
        return HtmlParser.find_forms(parsed_html, url), cookies

    @staticmethod
    def find_meta(parsed: dict or list) -> dict:
        """
        Search metadata inside a parsed html (dict)
        :param parsed: A parsed html
        :return: A dictionary of metadata tags like {'meta[name]': {'attr1': 'attr1_val' ...}}
        """
        return HtmlParser.__find_tags(parsed, HtmlParser._meta_tags)

    @staticmethod
    def find_forms(parsed: dict or list, url=None) -> dict:
        """
        Search forms inside parsed html (dict)
        :param parsed: A parsed html
        :param url: The parsed url (or None)
        :return: The list of found forms
        """
        form = dict()
        if parsed is None:
            return form
        if type(parsed) == dict:
            parsed_children = parsed.get('children')
            if 'form' == parsed.get('tag'):
                attrs = parsed.get('attrs')
                action = attrs.get('action')
                method = attrs.get('method')
                name = attrs.get('name')
                if action is None:
                    action = url
                if method is None:
                    method = HttpRequest.Type.POST
                attrs = {
                    'method': method,
                    'action': action,
                    'name': name
                }
                if name is None:
                    name = action
                form = {
                    'tag': 'form',
                    'attrs': attrs,
                    'name': name,
                    'children': HtmlParser.__find_inputs(parsed_children)
                }
            children = HtmlParser.find_forms(parsed_children, url)
            if children is not None and len(children) > 0:
                if len(form) > 0:
                    form['children'] = children
                else:
                    if len(children) == 1 and children.get('tag') is None:
                        form = children.get(0)
                    else:
                        form = children
        elif type(parsed) is list:
            children = dict()
            index = 0
            for value in parsed:
                child = HtmlParser.find_forms(value, url)
                if len(child) > 0:
                    children[index] = child
                    index += 1
            if len(children) == 1 and children.get(0).get('tag') is None:
                form = children.get(0)
            else:
                form = children
        else:
            Log.error(str(parsed) + ' is not a valid parsed content!')
        return form

    @staticmethod
    def find_links(parsed: dict or list) -> set:
        """
        Search links inside a parsed html (dict)
        :param parsed: A parsed html
        :return: A set of found links
        """
        links = set()
        if parsed is None:
            return links
        if type(parsed) is dict:
            attrs = parsed.get('attrs')
            if attrs is not None:
                url = None
                if 'form' == parsed.get('tag'):
                    url = attrs.get('action')
                elif 'a' == parsed.get('tag'):
                    url = attrs.get('href')
                    if url is not None and 'email-protection' in url:
                        # Skip email protections
                        url = None
                elif 'img' == parsed.get('tag'):
                    url = attrs.get('src')
                if url is not None:
                    links.add(url)
            links = links.union(HtmlParser.find_links(parsed.get('children')))
        elif type(parsed) is list:
            for value in parsed:
                links = links.union(HtmlParser.find_links(value))
        else:
            Log.error(str(parsed) + ' is not a valid parsed content!')
        return links

    @staticmethod
    def print_parsed(parsed: dict or list, depth: int = 0):
        """
        Print The result of methods @parse and @relevant_parse (so a parsed html)
        :param parsed: A parsed html
        :param depth: Current depth to build a pretty tree
        """
        space = ' ' * depth
        if type(parsed) == dict:
            print(space + '{')
            for key, value in parsed.items():
                if key == 'children':
                    HtmlParser.print_parsed(value, depth + 1)
                elif is_listable(value):
                    print((space + '  ') + str(key) + ':')
                    HtmlParser.print_parsed(value, depth + 2)
                else:
                    print((space + '  ') + str(key) + ': ' + str(value))
            print(space + '}')
        elif type(parsed) == list:
            for value in parsed:
                HtmlParser.print_parsed(value, depth + 1)
        else:
            Log.error(str(parsed) + ' is not a valid parsed content!')

    @staticmethod
    def types() -> tuple:
        return (
            HtmlParser.TYPE_ALL,
            HtmlParser.TYPE_RELEVANT,
            HtmlParser.TYPE_FORM,
            HtmlParser.TYPE_META
        )

    @staticmethod
    def type_descriptions() -> dict:
        return {
            HtmlParser.TYPE_ALL: 'Parse all tags',
            HtmlParser.TYPE_RELEVANT: 'Parse anchor, script, links, forms, images',
            HtmlParser.TYPE_FORM: 'Parse forms',
            HtmlParser.TYPE_META: 'Parse metadata'
        }

    @staticmethod
    def __abstract_parse(url: str, html: str, relevant: bool, cookies: str = None) -> (dict, str):
        """
        Make an HTML/URL parsing
        :param url: The url to parse (or None)
        :param html: The html page to parse as string (or None)
        :param relevant: True if should be processed only the tags in HtmlParse._relevant_tags
        :param cookies: The cookies to use on parsing
        :return: dictionary of tags, cookies
        """
        parser = HtmlParser(relevant)
        return parser.__parse(url, html, cookies)

    @staticmethod
    def __find_tags(parsed: dict or list, tags: dict) -> dict:
        """
        Search a certain kinds of tag inside a parsed html (dict)
        :param parsed: A parsed html
        :tags parsed: The tags to find
        :return: A dictionary of tags like {'tag[name]': {'attr1': 'attr1_val' ...}}
        """
        found_tag = {}
        if parsed is None:
            return found_tag
        if type(parsed) is dict:
            tag: str = parsed.get('tag')
            if tag is not None:
                tag = tag.lower()
            if tag in tags:
                attrs = parsed.get('attrs')
                # Tag
                found_tag = {'tag': tag}
                # Tag attrs
                found_tag_attrs = dict()
                for key, value in attrs.items():
                    found_tag_attrs[key] = value.strip()
                if HtmlParser._input_tags == tags and found_tag_attrs.get('type') is None:
                    found_tag_attrs['type'] = 'input'
                found_tag['attrs'] = found_tag_attrs
                # Tag data
                found_tag['data'] = parsed.get('data')
                # Tag name
                name = None
                for tag_name in HtmlParser._tag_names:
                    name = attrs.get(tag_name)
                    if name is not None:
                        break
                if name is None:
                    for tag_key in HtmlParser._tag_key_names:
                        name = attrs.get(tag_key)
                        if name is not None:
                            name = tag_key
                            break
                if name is None:
                    name = tag
                found_tag['name'] = name
            # Tag children
            children = HtmlParser.__find_tags(parsed.get('children'), tags)
            if children is not None and len(children) > 0:
                if len(found_tag) > 0:
                    found_tag['children'] = children
                else:
                    if len(children) == 1 and children.get(0).get('name') is None:
                        found_tag = children.get(0)
                    else:
                        found_tag = children
        elif type(parsed) is list:
            children = dict()
            index = 0
            for value in parsed:
                child = HtmlParser.__find_tags(value, tags)
                if len(child) > 0:
                    children[index] = child
                    index += 1
            if len(children) == 1 and children.get(0).get('name') is None:
                found_tag = children.get(0)
            else:
                found_tag = children
        else:
            Log.error(str(parsed) + ' is not a valid parsed content!')
        return found_tag

    @staticmethod
    def __find_inputs(parsed: dict or list) -> dict:
        """
        Search inputs inside a parsed html (dict)
        :param parsed: A parsed html
        :return: A dictionary of input tags like {'input[name]': {'attr1': 'attr1_val' ...}}
        """
        return HtmlParser.__find_tags(parsed, HtmlParser._input_tags)

    def handle_starttag(self, tag: str, attrs):
        """
        Start tag handler
        :param tag: The opened tag
        :param attrs: The attributes of opened tag like list[tuple{2}]
        """
        tag = str(tag).lower()
        if self.relevant and tag not in HtmlParser._relevant_tags.keys():
            self.queue_tag_ignored.append(tag)
            if tag in HtmlParser._not_closed_tags:
                self.handle_endtag(tag)
            return
        tag_attrs = {}
        for attr in attrs:
            attr_key = str(attr[0]).lower()
            attr_value = str(attr[1])
            if self.relevant and attr_key not in HtmlParser._relevant_tags.get(tag):
                continue
            if self.base_url is not None and attr_key in HtmlParser._url_attrs and (not is_url(attr_value)):
                if len(attr_value) == 0:
                    continue
                if attr_value[0] == '#':
                    continue
                if attr_value[0:2] == '//':
                    attr_value = self.url_scheme + ':' + attr_value
                else:
                    if attr_value[0] != '/' and self.base_url[len(self.base_url) - 1] != '/':
                        attr_value = '/' + attr_value
                    elif attr_value[0] == '/' and self.base_url[len(self.base_url) - 1] == '/':
                        attr_value = attr_value[1:]
                    attr_value = self.base_url + attr_value
            if 'email-protection' in attr_value:
                attr_value = 'email-protection'
            tag_attrs[attr_key] = attr_value
        cur_tag = {'tag': tag, 'attrs': tag_attrs}
        self.queue_tag.append(cur_tag)
        if tag in HtmlParser._not_closed_tags:
            self.handle_endtag(tag)

    def handle_endtag(self, tag: str):
        """
        Closing tag handler
        :param tag: The closing tag
        """
        tag = str(tag).lower()
        if len(self.queue_tag_ignored) > 0 and tag == self.queue_tag_ignored[-1]:
            self.queue_tag_ignored.pop()
            return
        if len(self.queue_tag) == 0:
            return
        cur_tag = self.queue_tag.pop()
        self.__nest_tag(cur_tag)

    def handle_data(self, data: str):
        """
        Tag content tag handler
        :param data: The last opened tag content
        """
        if len(self.queue_tag) == 0:
            return
        cur_tag = self.queue_tag[-1]
        if (not self.relevant) or 'data' in HtmlParser._relevant_tags.get(cur_tag.get('tag')):
            cur_tag['data'] = data.strip()
            self.queue_tag[-1] = cur_tag

    def __parse(self, url: str = None, html: str = None, cookies: str = None) -> (dict, str):
        """
        Make an HTML/URL parsing by processing ALL found tags
        :param url: The url to parse (or None)
        :param html: The html page to parse as string (or None)
        :param cookies: The cookies to use on parsing
        :return: dictionary of tags, cookies
        """
        self.url = None
        self.base_url = None
        is_image = False
        if url is not None:
            self.url = url
            url_parsed = urlparse(url)
            self.url_scheme = str(url_parsed.scheme)
            self.base_url = self.url_scheme + '://' + str(url_parsed.netloc)
            r = HttpRequest.request(url, cookies=cookies)
            if r is None:
                return None
            if r.status_code >= 400 or r.headers.get('Content-Type') in HtmlParser._unacceptable_content_types:
                return None
            try:
                html = r.json()
                Log.warning('Trying to parse a json with HTML parser!')
            except ValueError:
                html = r.text
            if r.headers is not None:
                for k, v in r.headers.items():
                    if k.lower() == 'set-cookie':
                        cookies = v
            if HttpRequest.is_image(r):
                is_image = True
                xmp_start = html.find('<x:xmpmeta')
                xmp_end = html.find('</x:xmpmeta')
                xmp_str = html[xmp_start:xmp_end+12]
                html = xmp_str
        if is_image:
            sorted_html = html
        else:
            sorted_html, errors = tidy_document(html)   # Sort html (and fix errors)
        self.feed(sorted_html)
        if cookies is None:
            cookies = ''
        return self.tags, cookies

    def __nest_tag(self, tag_dict: dict):
        """
        Insert the input tag inside the last one of queue
        :param tag_dict: A parsed-tag dictionary
        :return: void
        """
        queue_tag_len = len(self.queue_tag)
        if queue_tag_len == 0:
            parent = self.tags
        else:
            parent = self.queue_tag[-1]
        parent_children = parent.get('children')
        if parent_children is None:
            parent_children = []
        parent_children.append(tag_dict)
        parent['children'] = parent_children
        if queue_tag_len == 0:
            self.tags = parent
        else:
            self.queue_tag[-1] = parent



"""
*********************************************************************************
*                                                                               *
* abstract_sniffing_view.py -- Django Abstract Sniffing view.                   *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os

from black_widow.app.env import APP_STORAGE_OUT
from black_widow.app.helpers.storage import check_folder
from black_widow.app.helpers.util import now
from black_widow.app.managers.sniffer import PcapSniffer
from black_widow.app.services import JsonSerializer, MultiTask

from black_widow.app.gui.web.black_widow.models import SniffingJobModel
from black_widow.app.gui.web.black_widow.views.abstract_job_view import AbstractJobView


class AbstractSniffingView(AbstractJobView):
    """
    Abstract Sniffing View
    """
    storage_out_dir = os.path.join(APP_STORAGE_OUT, 'sniffing')
    check_folder(storage_out_dir)
    model_class = SniffingJobModel

    if not os.access(storage_out_dir, os.X_OK):
        os.chmod(storage_out_dir, 0o0755)

    def _new_job(self, filters: str, pcap: str, interfaces: list) -> SniffingJobModel:
        job = SniffingJobModel()
        job.filters = filters
        job.pcap_file = pcap
        job.interfaces = interfaces
        job.json_file = os.path.join(self.storage_out_dir, now() + '_SNIFFING_.json')

        def _sniffer_callback(pkt: dict):
            """
            The callback function of packet sniffer.
            This method writes the sniffed packets in a json file
            :param pkt: The sniffed packet
            """
            JsonSerializer.add_item_to_dict(pkt['number'], pkt, job.json_file)

        pcap_sniffer = PcapSniffer(
            filters=job.filters,
            src_file=job.pcap_file,
            interfaces=job.interfaces,
            limit_length=10000,
            callback=_sniffer_callback
        )

        job.pid_file = MultiTask.multiprocess(pcap_sniffer.start, asynchronous=True, cpu=1)
        job.save()

        return job

    def _copy_job(self, job: SniffingJobModel) -> SniffingJobModel:
        return self._new_job(
            job.filters,
            job.pcap_file,
            job.interfaces,
        )



# -*- coding: utf-8 -*-

"""
*********************************************************************************
*                                                                               *
* multitask.py -- Classes and methods to take advantage of multitasking.        *
*                                                                               *
* ---- Multi-processing ----                                                    *
*                                                                               *
* The multi-processing is very faster then multi-threading, but needs more      *
* hardware resources, because load in RAM multiple copy of the same part of the *
* program.                                                                      *
*                                                                               *
* ---- Multi-threading ----                                                     *
*                                                                               *
* The multi-threading is lower than multi-processing because forces the same    *
* process to perform multiple jobs simultaneously.                              *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import itertools
import multiprocessing
import os
import signal
import threading

from black_widow.app.env import APP_TMP
from black_widow.app.services import Log
from black_widow.app.helpers import storage
from black_widow.app.helpers.util import is_listable, timestamp


# Classe sfruttata sia da MultiThreading che da MultiProcessing
class MultiTask:
    # TODO: Implements serializer and deserializer (using "pickle")
    #       to get and write the result in file
    #       pickle implemented in: serializer.PickleSerializer

    CPU = multiprocessing.cpu_count()
    MULTI_THREADING = 'MULTI_THREADING'
    MULTI_PROCESSING = 'MULTI_PROCESSING'
    tasks_types = (MULTI_THREADING, MULTI_PROCESSING)

    @staticmethod
    def get_pids_from_file(pidfile: str) -> list:
        """
        Returns a list of pids inside the pidfile
        :type pidfile str
        :param pidfile The pidfile returned by MultiTask.multiprocess()
        """
        if pidfile is None:
            return []
        return storage.read_file(pidfile).split(', ')

    # Crea un sottoprocesso che a sua volta genera n(=cpu) threads.
    # La creazione del sottoprocesso è essenziale, in quanto il primo thread che
    # riesce a trovare la soluzione, killerà il suo stesso processo, per interrompere
    # anche tutti gli altri threads ed evitare attesa ed utilizzo di risorse inutile.
    # Sfrutta il multithreading per effettuare la stessa operazione sugli elementi
    # di un/una lista|tupla|dizionario|range.
    # Si assume che tutti gli argomenti di tipo lista|tupla|dizionario|range
    # debbano essere smistati ai vari threads
    # @param target La funzione che i threads chiameranno: se questa ritorna un valore
    #               diverso da None, allora tutti i threads concorrenti verranno stoppati
    # @param args Gli argomenti che verranno passati alla funzione target
    # @param asynchronous True, se NON bisogna attendere la fine dell'esecuzione di
    #                     tutti i threads, False altrimenti
    # @param cpu Il numero di threads da creare (default: il numero di cpu disponibili)
    # @return Il risultato della funzione target
    @staticmethod
    def multithread(target=None, args=(), asynchronous=False, cpu=CPU):
        multitask = MultiTask(MultiTask.MULTI_THREADING)

        if cpu == 1:
            # If only one thread is required, the parent process is not needed
            return multitask.start(target, args, asynchronous, cpu)

        # The arguments to pass at the parent process
        multithread_args = (target, args, asynchronous, cpu)
        # Creates a process (cpu=1) that runs all the threads
        return MultiTask.multiprocess(multitask.start, multithread_args, cpu=1)

    # Sfrutta il multiprocessing per effettuare la stessa operazione
    # sugli elementi di un/una lista|tupla|dizionario|range.
    # Si assume che tutti gli argomenti di tipo lista|tupla|dizionario|range
    # debbano essere smistati ai vari processi
    # @param target La funzione che i processi chiameranno: se questa ritorna un valore
    #               diverso da None, allora tutti i processi concorrenti verranno stoppati
    # @param args Gli argomenti che verranno passati alla funzione target
    # @param asynchronous True, se non bisogna attendere la fine dell'esecuzione di
    #                     tutti i processi, False altrimenti
    # @param cpu Il numero di processi da creare (default: il numero di cpu disponibili)
    # @return Il risultato della funzione target se non async, il pidfile altrimenti
    @staticmethod
    def multiprocess(target=None, args=(), asynchronous=False, cpu=CPU):
        multitask = MultiTask(MultiTask.MULTI_PROCESSING)
        result = multitask.start(target, args, asynchronous, cpu)
        if asynchronous:
            return multitask.pidfile
        return result

    # @param tasks_type in MultiTask.tasks_types
    def __init__(self, tasks_type):
        if tasks_type not in MultiTask.tasks_types:
            tasks_type = MultiTask.MULTI_PROCESSING
            Log.error(str(tasks_type) + ' is not a valid tasks type!')
        self.tasks_type = tasks_type
        if self.tasks_type == MultiTask.MULTI_PROCESSING:
            self.Multitask = multiprocessing.Process
            self.tag = 'Process '
        else:
            self.Multitask = threading.Thread
            self.tag = 'Thread '
        self.tasks = []
        pid = str(multiprocessing.process.current_process().pid)
        # File con pids (se multiprocessing)
        self.pidfile = APP_TMP + '/multitask.' + pid + '.' + timestamp() + '.pids'
        # File con result
        self.resfile = APP_TMP + '/multitask.' + pid + '.' + timestamp() + '.res'

    # Sfrutta il multiprocessing o il multitasking per effettuare la stessa
    # operazione sugli elementi di un/una lista|tupla|dizionario|range.
    # Si assume che tutti gli argomenti di tipo lista|tupla|dizionario|range
    # debbano essere smistati ai vari processi
    # @param target La funzione che i processi chiameranno
    # @param args Gli argomenti che verranno passati alla funzione
    # @param asynchronous True, se non bisogna attendere la fine dell'
    #                     esecuzione di tutti i tasks, False altrimenti
    # @param cpu Il numero di cpu da usare
    def start(self, target, args, asynchronous, cpu):
        self.tasks = []

        def task_target(*arguments):
            result = None
            if self.tasks_type == MultiTask.MULTI_PROCESSING:
                curr_task = multiprocessing.process.current_process()
                Log.info(self.tag + 'started (PID=' + str(curr_task.pid) + ')')
            else:
                curr_task = threading.current_thread()
                Log.info(self.tag + 'started')
            if target is not None:
                result = target(*arguments)
            if result is not None:
                Log.success("Result: " + str(result))
                # Scrivo il risultato nel file
                Log.info('Writing result in ' + str(self.resfile))
                storage.overwrite_file(str(result), self.resfile)   # TODO: dump result as object with "pickle"
                # Termino tutti gli altri threads/processi
                if self.tasks_type == MultiTask.MULTI_PROCESSING:
                    Log.info('Killing other processes')
                    running_pids = MultiTask.get_pids_from_file(self.pidfile)
                    for pid in running_pids:
                        pid = int(pid)
                        if pid == curr_task.pid:
                            continue
                        try:
                            os.kill(pid, signal.SIGKILL)
                            Log.info('Process ' + str(pid) + ' killed!')
                        except Exception as ex:
                            Log.error(str(ex))
                    Log.info(self.tag + 'end')
                else:
                    Log.info('Ignoring other threads')
                    # Killa se stesso
                    pid = multiprocessing.process.current_process().pid
                    Log.info(self.tag + 'end')
                    os.kill(pid, signal.SIGKILL)

        for i in range(0, cpu):
            task_args = ()
            for arg in args:
                Log.info('Argument type: ' + str(type(arg)))
                if is_listable(arg):
                    # Divido gli elementi in 1/cpu parti
                    p_list_len = (len(arg) / cpu) + (len(arg) % cpu)
                    if type(arg) == dict:
                        iterator = iter(arg.items())
                        task_args += (
                            dict(itertools.islice(iterator, int((i * p_list_len)), int((i + 1) * p_list_len))),
                        )
                    else:
                        task_args += (arg[int((i * p_list_len)):int(((i + 1) * p_list_len))],)
                else:
                    task_args += (arg,)
            task = self.Multitask(target=task_target, args=task_args)
            self.tasks.append(task)

        if self.tasks_type == MultiTask.MULTI_PROCESSING:
            pids = []
            try:
                signal.signal(signal.SIGCHLD, signal.SIG_IGN)   # Ignore child exit status
            except ValueError as e:
                # Probably you have executed Django
                Log.error(str(e))
            for task in self.tasks:
                task.start()
                # noinspection PyUnresolvedReferences
                pids.append(task.pid)
            storage.overwrite_file(str(pids).strip('[]'), self.pidfile)
        else:
            for task in self.tasks:
                task.start()

        if not asynchronous:
            # Attende la fine dell'esecuzione di tutti i tasks
            for task in self.tasks:
                task.join()
                Log.info('Task ' + str(task.name) + ' joined')
            Log.info('Reading result in ' + str(self.resfile))
            # Prendo il risultato dal file
            res = storage.read_file(self.resfile)   # TODO: load result as object with "pickle"
            # Elimino l'eventuale file con i pid
            storage.delete(self.pidfile)
            # Elimino il file con il risultato
            storage.delete(self.resfile)
            Log.success('MultiTask -> result: ' + str(res))
            return res

        return None



"""
*********************************************************************************
*                                                                               *
* abstract_job_view.py -- Django Abstract Job view                              *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import signal

from django.core.exceptions import ObjectDoesNotExist
from django.http import JsonResponse
from django.shortcuts import redirect, render

from black_widow.app.gui.web.black_widow.models.abstract_job_model import AbstractJobModel
from black_widow.app.gui.web.black_widow.views.abstract_view import AbstractView
from black_widow.app.services import Log


class AbstractJobView(AbstractView):
    model_class: AbstractJobModel = None

    """
    Abstract Job View
    """
    def _get_job(self, request, redirect_url: str):
        """
        Show the requested job
        :type request: django.core.handlers.wsgi.WSGIRequest
        :param redirect_url: The url to redirect the request in case of errors
        :return: django.http.HttpResponse
        """
        request_params: dict = request.GET.dict()

        try:
            job_id = int(request_params.get('id'))
        except (ValueError, TypeError) as e:
            Log.error(str(e))
            return redirect(redirect_url)

        Log.info("Showing job #" + str(job_id))

        job = self._get_job_model(job_id)
        if job is None:
            return redirect(redirect_url)

        return render(request, self.template_name, {
            'job': job
        })

    def _post_job(self, request) -> JsonResponse:
        """
        :type request: django.core.handlers.wsgi.WSGIRequest
        :return: django.http.JsonResponse
        """
        # noinspection PyTypeChecker
        job: AbstractJobModel = None
        request_params: dict = request.POST.dict()
        job_id = request_params.get('id')
        try:
            job_id = int(job_id)
            job = self._get_job_model(job_id)
        except ValueError:
            pass
        except Exception as e:
            Log.error(str(e))

        if job is None:
            return JsonResponse({
                'message': 'Unable to find the requested job'
            }, status=400)

        signal_job = request_params.get('signal')
        if signal_job is not None:
            signal_job = int(signal_job)

            if signal_job == 0:   # Custom signal 0 = Restart capturing
                job_new = self._copy_job(job)
                job_id = job_new.id
                signal_job = signal.SIGABRT

            try:
                job.kill(signal_job)
            except ProcessLookupError:
                Log.warning("The process " + str(job.pid) + " does not exists")

            if signal_job == signal.SIGABRT:    # 6 = Abort permanently by cleaning job
                if not job.delete():
                    return JsonResponse({
                        'message': 'Unable to delete the job'
                    }, status=400)

            return JsonResponse({
                'id': job_id,
                'signal': signal_job,
                'message': 'Signal sent'
            }, status=200)

        job.self_check()

        page = request_params.get('page')
        page_size = request_params.get('page_size')
        pagination = self.pagination(job.json_dict, page, page_size)
        pagination.update({
            'job': {
                'id': job_id,
                'status': job.status_name
            }
        })

        return JsonResponse(pagination, status=200)

    def _copy_job(self, job: AbstractJobModel) -> AbstractJobModel:
        raise NotImplementedError("Abstract method _copy_job")

    def _get_job_model(self, job_id: int) -> AbstractJobModel or None:
        """
        Get the job with id job_id
        :param job_id: An id of job
        :return: The searched job, or None
        """
        try:
            job = self.model_class.objects.get(id=job_id)
        except ObjectDoesNotExist:
            return None
        return job



# -*- coding: utf-8 -*-

"""
*********************************************************************************
*                                                                               *
* logger.py -- The logger class for global usage.                               *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import calendar
import datetime
import os

from multiprocessing.process import current_process
from threading import current_thread
from termcolor import colored

from black_widow.app.env import APP_LOGFILE, APP_PROC, APP_DEBUG
from black_widow.app.helpers import storage


# Logger
class Log:
    logger = None

    @staticmethod
    def info(msg):
        Log._log('inf', msg)

    @staticmethod
    def success(msg):
        Log._log('suc', msg)

    @staticmethod
    def error(msg):
        Log._log('err', msg)

    @staticmethod
    def warning(msg):
        Log._log('war', msg)

    @staticmethod
    def _log(method: str, msg):
        if Log.logger is None:
            Log.logger = Log()
        getattr(Log.logger, '_' + method)(str(msg))

    def __init__(self):
        storage.touch(APP_LOGFILE)
        try:
            storage.chmod(APP_LOGFILE, 0o0666)
        except PermissionError:
            pass

    # info
    def _inf(self, msg):
        self.__log__(msg, 'INFO', 'cyan')

    # error
    def _err(self, msg):
        self.__log__(msg, 'ERROR', 'red')

    # warning
    def _war(self, msg):
        self.__log__(msg, 'WARNING', 'yellow')

    # success
    def _suc(self, msg):
        self.__log__(msg, 'DONE', 'green')

    def __log__(self, msg, _type, color):
        self._check_log_file()
        msg_log = self._get_log_header() + ' ' + colored(_type, color)
        curr_process = current_process()
        curr_thread = current_thread()
        if 'Main' not in curr_process.name:
            msg_log += ' [' + curr_process.name + ']'
        if 'Main' not in curr_thread.name:
            msg_log += ' (' + curr_thread.name + ')'
        msg_log += ': ' + msg.rstrip()
        if APP_DEBUG:
            print(msg_log)
        storage.append_in_file(msg_log, APP_LOGFILE)

    def _get_log_header(self):
        return self._get_timestamp() + ' ' + APP_PROC

    @staticmethod
    def _check_log_file():
        storage.check_folder(os.path.dirname(APP_LOGFILE))
        # Se il file di log pesa almeno 5 MB, lo sovrascrivo
        if os.path.isfile(APP_LOGFILE):
            try:
                size_file_mb = os.stat(APP_LOGFILE).st_size / 1000000.0
                if size_file_mb >= 5:
                    storage.delete(APP_LOGFILE)
            except FileNotFoundError:
                # Un altro thread o processo ha già eliminato il file
                return

    @staticmethod
    def _get_timestamp():
        now_month = int(datetime.datetime.now().strftime("%m"))
        timestamp = calendar.month_abbr[now_month] + ' ' + datetime.datetime.now().strftime("%d %H:%M:%S")
        return timestamp



# -*- coding: utf-8 -*-

"""
*********************************************************************************
*                                                                               *
* mac_manufacturer.py -- Mac Manufacturer Lookup class                          *
*                                                                               *
*   MAC Address explanation:                                                    *
*   _________________________                                                   *
*   | 1 | 2 | 3 | 4 | 5 | 6 |  =  48bit (6 * 8bit)                              *
*   {^^^^OUI^^^}^{^^^NIC^^^^}                                                   *
*                                                                               *
*   MAC Address (6 Octets):                                                     *
*       |--- First 3 Octets:  OUI "Organisationally Unique Identifier"          *
*       |       |--- 1st Octet:                                                 *
*       |              |--- 7th bit (1=locally administered, 0=globally unique) *
*       |              |--- 8th bit (0=unicast, 1=multicast)                    *
*       |--- Second 3 Octets: NIC "Network Interface Controller"                *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from math import floor
from os.path import join

from black_widow.app.env import APP_STORAGE
from black_widow.app.managers.request import HttpRequest
from black_widow.app.services import Log, JsonSerializer


class MacManufacturer:
    """
    The Mac Manufacturer Lookup class
    """

    MANUFACTURERS_DOMAIN = 'code.wireshark.org'
    
    _MANUFACTURERS_URL = 'https://' + MANUFACTURERS_DOMAIN + '/review/gitweb?p=wireshark.git;a=blob_plain;f=manuf'
    _MANUFACTURERS_DETAIL_DICT = {
        0: 'mac',
        1: 'vendor',
        2: 'company',
        3: 'comment'
    }
    _MANUFACTURERS_JSON = join(APP_STORAGE, 'mac_manufacturers.json')

    _mac_manufacturer = None

    def __init__(self):
        self._manufacturer_dict = JsonSerializer.get_dictionary(MacManufacturer._MANUFACTURERS_JSON)
        self._update_manufacturer_dict()

    @staticmethod
    def lookup(mac: str) -> dict:
        """
        Lookup the mac address
        :param mac: The mac address to lookup
        :return: The looked up mac
        """
        if MacManufacturer._mac_manufacturer is None:
            MacManufacturer._mac_manufacturer = MacManufacturer()
        mac_manufacturer_result = MacManufacturer._mac_manufacturer._lookup(mac)
        if type(mac_manufacturer_result) is not dict:
            return dict()
        mac_manufacturer_result.update({
            'mac_full': mac,
            'manufacturer': mac_manufacturer_result.get('vendor') + '_' + mac[9:]
        })
        return mac_manufacturer_result

    def _lookup(self, mac: str) -> dict or None:
        """
        Lookup the mac address
        :param mac: The mac address to lookup
        :rtype: dict or None
        """
        for i in range(len(mac), 0, -1):
            manufacturer = self._manufacturer_dict.get(mac[0:i])
            if manufacturer is not None:
                return manufacturer
        return None

    def _update_manufacturer_dict(self):
        manufacturer_response = HttpRequest.request(MacManufacturer._MANUFACTURERS_URL, timeout=10)
        if manufacturer_response is None:
            return
        if manufacturer_response.text is None:
            return
        manufacturer_dict = dict()
        manufacturer_list = manufacturer_response.text.splitlines()
        for manufacturer in manufacturer_list:
            if len(manufacturer) < 1:
                continue
            if manufacturer[0] == '#':
                continue
            manufacturer_details = manufacturer.split('\t')
            i = 0
            mac = None
            lookup_dict = {
                MacManufacturer._MANUFACTURERS_DETAIL_DICT[1]: None,
                MacManufacturer._MANUFACTURERS_DETAIL_DICT[2]: None,
                MacManufacturer._MANUFACTURERS_DETAIL_DICT[3]: None
            }
            for detail in manufacturer_details:
                if detail == '':
                    continue
                if i == 0:
                    # MAC address
                    mac_detail = detail.split('/')
                    if len(mac_detail) == 2:
                        # The mac has a sub mask, so the dict key is the first n bits
                        sub_mask = int(mac_detail[1]) / 4
                        mac_sub_mask = floor(sub_mask + (sub_mask / 2))
                        mac = mac_detail[0][0:mac_sub_mask]
                    elif len(mac_detail) == 1:
                        # The mac has not a sub mask
                        mac = mac_detail[0]
                    else:
                        Log.error("Wrong mac address: " + str(detail))
                        break
                if i >= len(MacManufacturer._MANUFACTURERS_DETAIL_DICT):
                    Log.error("Wrong manufacturer details: " + str(manufacturer_details))
                    break
                lookup_dict[MacManufacturer._MANUFACTURERS_DETAIL_DICT[i]] = detail
                i += 1
            if mac is None:
                Log.error("Wrong manufacturer details: " + str(manufacturer_details))
                continue
            manufacturer_dict[mac] = lookup_dict
        if len(manufacturer_dict) > 0:
            self._manufacturer_dict = manufacturer_dict
            JsonSerializer.set_dictionary(self._manufacturer_dict, MacManufacturer._MANUFACTURERS_JSON)



"""
*********************************************************************************
*                                                                               *
* sniffing_job_model.py -- A sniffing job info model                            *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.db import models

from black_widow.app.gui.web.black_widow.models.abstract_job_model import AbstractJobModel
from black_widow.app.helpers import storage


class SniffingJobModel(AbstractJobModel):
    """
    Django Sniffing Job Model
    """
    filters: str = models.TextField(null=True)
    _interfaces: str or None = models.TextField(null=False)
    pcap_file: str = models.CharField(max_length=250, null=True)
    json_sort_value = 'number'

    @staticmethod
    def all() -> models.query.QuerySet:
        return AbstractJobModel._all(SniffingJobModel)

    @property
    def interfaces(self) -> list or None:
        if self._interfaces is None:
            return None
        return self._interfaces.split(';')

    @interfaces.setter
    def interfaces(self, value: list or None):
        if value is None:
            self._interfaces = None
        else:
            self._interfaces = ';'.join(value)

    def delete(self, using=None, keep_parents=False):
        if not storage.delete(self.pcap_file):
            return False
        return super(SniffingJobModel, self).delete(using, keep_parents)

    def __str__(self) -> str:
        return 'SniffingJobModel(' + str({
            'id': self.id,
            'filters': self.filters,
            'interfaces': self.interfaces,
            'json_file': self.json_file,
            'pcap_file': self.pcap_file,
            'pid': self.pid,
            'pid_file': self.pid_file,
            'status': self.status_name
        }) + ')'



"""
*********************************************************************************
*                                                                               *
* abstract_model.py -- The black-widow AbstractModel.                           *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.db import models


class AbstractModel(models.Model):
    """
    Abstract Model
    """
    objects: models.manager.Manager
    id: int = models.BigAutoField(primary_key=True)

    class Meta:
        abstract = True



"""
*********************************************************************************
*                                                                               *
* util.py -- Generic useful methods, used in all parts of software.             *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import getpass
import random
import re
import os
import string
import subprocess

from datetime import datetime

from black_widow.app.env import IGNORE_NON_ROOT


def now() -> str:
    """
    :return: The current datetime as string
    """
    return str(datetime.now()).replace(' ', '_')


def timestamp() -> str:
    """
    :return: The current timestamp as string
    """
    return str(datetime.timestamp(datetime.now()))


def regex_in_string(regex, text) -> bool:
    """
    Checks if the input string contains the input regex
    :param regex: The regex to search
    :param text: The string to check
    :return: True if string contains regex, otherwise False
    """
    reg = re.compile(regex)
    matches = re.findall(reg, text)
    return len(matches) > 0


def regex_is_string(regex, text) -> bool:
    """
    Check if a string is equals to input regex
    :param regex: The regex to compare
    :param text: The string to compare
    :return: True, if the regex is equals to string
    """
    reg = re.compile(regex)
    return bool(reg.match(text))


def replace_regex(regex, replace, text) -> str:
    """
    Replace the regex in the string with another string
    :param regex: The regex to found
    :param replace: The replacer string
    :param text: The string where find and replace the regex
    :return: The input string with replaced regex
    """
    return re.sub(regex, replace, text, flags=re.M)


def is_listable(obj) -> bool:
    """
    Check if type(obj) is on of (list, tuple, dict, range)
    :param obj: An Object
    :return: True if obj is listable, otherwise False
    """
    return type(obj) in (list, tuple, dict, range)


def sort_dict(dictionary: dict) -> dict:
    """
    :param dictionary: The dictionary to sort
    :return: The sorted dictionary
    """
    sorted_dictionary = dict()
    i = 0
    for k in dictionary.keys():
        sorted_dictionary[i] = dictionary[k]
        i += 1
    return sorted_dictionary


def reverse_dict(dictionary: dict) -> dict:
    """
    :param dictionary: The dictionary to sort
    :return: The sorted dictionary
    """
    reversed_dictionary = dict()
    keys = list(dictionary.keys())
    keys.reverse()
    for k in keys:
        reversed_dictionary[k] = dictionary[k]
    return reversed_dictionary


def print_dict(dictionary: dict, depth: int = 0):
    """
    :type dictionary: dict
    :type depth: int
    """
    for key, value in dictionary.items():
        space = (' ' * depth) + '|-- '
        print(space + str(key) + ':')
        if type(value) is dict:
            print_dict(value, depth+1)
        else:
            print(space + ' [' + str(type(value)) + '] ' + str(value))


def rand_str(special: bool = True, length: int = 10) -> str:
    """
    Generates a random string
    :param special: True if you want to include special chars, otherwise False
    :param length: The generated string length
    :return: A random string
    """
    if length <= 1:
        return ''
    chars = string.ascii_letters + string.digits + string.punctuation
    if special:
        chars += string.punctuation
    return ''.join(random.choice(chars) for _ in range(length))


#################
# Process Utils #
#################

def pid_exists(pid: int) -> bool:
    """
    Checks if exists a process with input pid
    :param pid: The pid to check
    :return: True, if the pid exists, otherwise False
    """
    if pid is None:
        return False
    try:
        os.kill(pid, 0)
    except OSError:
        return False
    return True


def is_executable(cmd_file: str) -> bool:
    """
    If cmd_file is a file this method checks if the current user can executes that,
    otherwise it checks if the current user can executes the associated file
    (returned by `type cmd_file`)
    :param cmd_file: The command or the file to executes
    :return: True, if the current user can executes "cmd_file", otherwise False
    """
    path = exec_type(cmd_file)
    return os.access(path, os.X_OK)


def is_root() -> bool:
    """
    If IGNORE_NON_ROOT in .env is False, this method checks if the current user is root
    :return: True, if the current user is root, otherwise False
    """
    if IGNORE_NON_ROOT:
        return True
    user = whoami(False)
    return user['uid'] == 0 or user['sudo']


def root_required():
    """
    If the current process has not root privileges, print the error and exits from app
    """
    if not is_root():
        print("\nThis feature requires root privileges!\n")
        exit(20)


def whoami(check_sudo: bool = True) -> dict:
    """
    Get the dictionary of current user info
    :param check_sudo: True, if you want get the user which executed sudo command
    :return: A dictionary with current user info
    """
    name = None
    uid = None
    gid = None
    sudo = True
    if check_sudo:
        name = os.environ.get('SUDO_USER')
        uid = os.environ.get('SUDO_UID')
        gid = os.environ.get('SUDO_GID')
    if name is None:
        sudo = False
        name = getpass.getuser()
        uid = os.getuid()
        gid = os.getgid()
    return {
        'name': str(name),
        'uid': int(uid),
        'gid': int(gid),
        'sudo': sudo
    }


def set_owner_process(user: dict):
    """ set user and group of workers processes """
    os.setgid(user['gid'])
    os.setuid(user['uid'])
    os.environ['USER'] = user['name']
    os.environ['LOGNAME'] = user['name']
    home_dirname = os.path.dirname(os.path.expanduser("~"))
    if home_dirname == '/' and user['sudo']:
        home_dirname = '/home'
    new_home = os.path.join(home_dirname, user['name'])
    os.environ['HOME'] = new_home


def os_exec(*args) -> list:
    """
    Executes through os stdin the input commands
    :param args: cmd + args (eg1. "cmd" "[args]"), (eg2. "ls" "-l")
    :return: The stdout of executed command
    """
    p = subprocess.Popen(args, shell=True, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
    list_stdout = []
    for line in p.stdout.readlines():
        list_stdout.append(str(line.decode('utf-8')).rstrip('\n'))
    return list_stdout


def exec_type(cmd: str) -> str or None:
    """
    Indicate how the input "cmd" would be interpreted if used as a command name.
    If "cmd" is a shell built-in or it does not exists, returns None
    :param cmd: The command to check
    :return: Usually, the interpreted path of input command
    """
    if type(cmd) is not str:
        raise ValueError("cmd must be a string")
    outputs = os_exec('type ' + cmd)
    if len(outputs) == 0:
        return None
    output: str = outputs[0]
    output_split = output.split(' ')
    cmd_file = output_split[-1]
    if os.path.isfile(cmd_file):
        return cmd_file
    return None



"""
*********************************************************************************
*                                                                               *
* network.py -- Useful methods to take network info of current device.          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import netifaces as ni
import socket

from contextlib import closing


def get_interfaces():
    """
    Get all network interfaces
    :rtype: list
    """
    return ni.interfaces()


def get_ip_address():
    """
    Get the current LAN ip address of this device
    :rtype: str
    """
    ip = None
    interfaces = get_interfaces()
    for interface in interfaces:
        try:
            ip = ni.ifaddresses(interface)[ni.AF_INET][0]['addr']
            break
        except (KeyError, Exception):
            pass
    return ip


def check_socket(host: str, port: int):
    """
    Check if socket is running, so if the port is open in chosen host
    :param host: The host   (eg. 127.0.0.1)
    :param port: The port   (eg. 443)
    :rtype: bool
    """
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        return sock.connect_ex((host, port)) == 0



"""
*********************************************************************************
*                                                                               *
* index_view.py -- Django index and static views.                               *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os

from django.http import HttpResponseNotFound, FileResponse
from django.shortcuts import render

from black_widow.app.gui.web.settings import STATICFILES_DIRS


def index(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :return: django.http.HttpResponse
    """
    return render(request, 'index.html')


def user(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: django.http.HttpResponse
    """
    return render(request, 'user.html')


def tables(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: django.http.HttpResponse
    """
    return render(request, 'tables.html')


def typography(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: django.http.HttpResponse
    """
    return render(request, 'typography.html')


def icons(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: django.http.HttpResponse
    """
    return render(request, 'icons.html')


def notifications(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: django.http.HttpResponse
    """
    return render(request, 'notifications.html')


def upgrade(request):
    """
    :type request: django.core.handlers.wsgi.WSGIRequest
    :rtype: django.http.HttpResponse
    """
    return render(request, 'upgrade.html')


# noinspection PyUnusedLocal
def static(request, path):
    """
    Manage requested static file (for non-DEBUG mode compatibility without web-server)
    :type request: django.core.handlers.wsgi.WSGIRequest
    :type path: str
    :rtype: django.http.HttpResponse
    """
    for directory in STATICFILES_DIRS:
        static_file = os.path.join(directory, path)
        if os.path.isfile(static_file):
            return FileResponse(open(static_file, 'rb'))
    return HttpResponseNotFound()



"""
*********************************************************************************
*                                                                               *
* update.py -- Methods to update the json file ``APP_SETTINGS``.                *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from black_widow.app.env import APP_SETTINGS, APP_DEBUG   # Json file
from black_widow.app.services import Log, JsonSerializer
from black_widow.app.helpers import validators

from . import keys


class Set:
    @staticmethod
    def my_ip(ip):
        if APP_DEBUG:
            Log.info('CALLED: Set.my_ip(' + str(ip) + ')')
        if not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Set.__set__(keys.MY_IP, ip)

    @staticmethod
    def my_interface(interface):
        if APP_DEBUG:
            Log.info('CALLED: Set.my_interface(' + str(interface) + ')')
        return Set.__set__(keys.MY_INTERFACE, interface)

    @staticmethod
    def team_token(token):
        if APP_DEBUG:
            Log.info('CALLED: Set.team_token(' + str(token) + ')')
        return Set.__set__(keys.TEAM_TOKEN, token)

    @staticmethod
    def submit_url(url):
        if APP_DEBUG:
            Log.info('CALLED: Set.submit_url(' + str(url) + ')')
        if not validators.is_url(url):
            Log.error(str(url) + ' is not a valid url')
            return False
        return Set.__set__(keys.SUBMIT_URL, url)

    @staticmethod
    def game_server(ip):
        if APP_DEBUG:
            Log.info('CALLED: Set.game_server(' + str(ip) + ')')
        if not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Set.__set__(keys.GAME_SERVER, ip)

    @staticmethod
    def flag_regex(regex):
        if APP_DEBUG:
            Log.info('CALLED: Set.flag_regex(' + str(regex) + ')')
        return Set.__set__(keys.FLAG_REGEX, regex)

    @staticmethod
    def __set__(key, value):
        dictionary = JsonSerializer.get_dictionary(APP_SETTINGS)
        dictionary[key] = value
        JsonSerializer.set_dictionary(dictionary, APP_SETTINGS)
        return True


# Add - Per gestire parametri di tipo lista
class Add:
    @staticmethod
    def server_to_attack(ip):
        if APP_DEBUG:
            Log.info('CALLED: Add.server_to_attack(' + str(ip) + ')')
        if not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Add.__add__(keys.SERVER_TO_ATTACK, ip)

    @staticmethod
    def server_to_defend(ip):
        if APP_DEBUG:
            Log.info('CALLED: Add.server_to_defend(' + str(ip) + ')')
        if not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Add.__add__(keys.SERVER_TO_DEFEND, ip)

    @staticmethod
    def team_player(ip):
        if APP_DEBUG:
            Log.info('CALLED: Add.team_player(' + str(ip) + ')')
        if not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Add.__add__(keys.TEAM_PLAYER, ip)

    # Aggiunge un elemento da una lista
    @staticmethod
    def __add__(key, element):
        dictionary = JsonSerializer.get_dictionary(APP_SETTINGS)
        elements = dictionary.get(key)
        if type(elements) != list:
            elements = []
        if element not in elements:
            elements.append(element)
            dictionary[key] = elements
            JsonSerializer.set_dictionary(dictionary, APP_SETTINGS)
        return True


class Remove:
    @staticmethod
    def my_ip():
        if APP_DEBUG:
            Log.info('CALLED: Remove.my_ip()')
        return Set.__set__(keys.MY_IP, None)

    @staticmethod
    def my_interface():
        if APP_DEBUG:
            Log.info('CALLED: Remove.my_interface()')
        return Set.__set__(keys.MY_INTERFACE, None)

    @staticmethod
    def team_token():
        if APP_DEBUG:
            Log.info('CALLED: Remove.team_token()')
        return Set.__set__(keys.TEAM_TOKEN, None)

    @staticmethod
    def game_server():
        if APP_DEBUG:
            Log.info('CALLED: Remove.game_server()')
        return Set.__set__(keys.GAME_SERVER, None)

    @staticmethod
    def submit_url():
        if APP_DEBUG:
            Log.info('CALLED: Remove.submit_url()')
        return Set.__set__(keys.SUBMIT_URL, None)

    @staticmethod
    def flag_regex():
        if APP_DEBUG:
            Log.info('CALLED: Remove.flag_regex()')
        return Set.__set__(keys.FLAG_REGEX, None)

    @staticmethod
    def server_to_attack(ip='*'):
        if APP_DEBUG:
            Log.info('CALLED: Remove.server_to_attack(' + str(ip) + ')')
        if ip != '*' and not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Remove.__remove__(keys.SERVER_TO_ATTACK, ip)

    @staticmethod
    def server_to_defend(ip='*'):
        if APP_DEBUG:
            Log.info('CALLED: Remove.server_to_defend(' + str(ip) + ')')
        if ip != '*' and not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Remove.__remove__(keys.SERVER_TO_DEFEND, ip)

    @staticmethod
    def team_player(ip='*'):
        if APP_DEBUG:
            Log.info('CALLED: Remove.team_player(' + str(ip) + ')')
        if ip != '*' and not validators.is_ip(ip):
            Log.error(str(ip) + ' is not a valid ip address')
            return False
        return Remove.__remove__(keys.TEAM_PLAYER, ip)

    # Rimuove un elemento da una lista
    @staticmethod
    def __remove__(key, element):
        dictionary = JsonSerializer.get_dictionary(APP_SETTINGS)
        if dictionary.get(key) is None:
            return True
        if element == '*':
            # Rimuove tutti gli elementi
            dictionary[key] = []
            JsonSerializer.set_dictionary(dictionary, APP_SETTINGS)
            return True
        elements = dictionary[key]
        if element in elements:
            elements.remove(element)
            dictionary[key] = elements
            JsonSerializer.set_dictionary(dictionary, APP_SETTINGS)
        return True


class Get:
    @staticmethod
    def all():
        if APP_DEBUG:
            Log.info('CALLED: Get.all()')
        return Get.__get__()

    @staticmethod
    def my_ip():
        if APP_DEBUG:
            Log.info('CALLED: Get.my_ip()')
        return Get.__get__(keys.MY_IP)

    @staticmethod
    def my_interface():
        if APP_DEBUG:
            Log.info('CALLED: Get.my_interface()')
        return Get.__get__(keys.MY_INTERFACE)

    @staticmethod
    def team_token():
        if APP_DEBUG:
            Log.info('CALLED: Get.team_token()')
        return Get.__get__(keys.TEAM_TOKEN)

    @staticmethod
    def game_server():
        if APP_DEBUG:
            Log.info('CALLED: Get.game_server()')
        return Get.__get__(keys.GAME_SERVER)

    @staticmethod
    def submit_url():
        if APP_DEBUG:
            Log.info('CALLED: Get.submit_url()')
        return Get.__get__(keys.SUBMIT_URL)

    @staticmethod
    def server_to_attack():
        if APP_DEBUG:
            Log.info('CALLED: Get.server_to_attack()')
        return Get.__get__(keys.SERVER_TO_ATTACK)

    @staticmethod
    def flag_regex():
        if APP_DEBUG:
            Log.info('CALLED: Get.flag_regex()')
        return Get.__get__(keys.FLAG_REGEX)

    @staticmethod
    def __get__(key=None):
        dictionary = JsonSerializer.get_dictionary(APP_SETTINGS)
        if key is None:
            return dictionary
        return dictionary.get(key)



"""
*********************************************************************************
*                                                                               *
* urls.py -- Web URL Configuration.                                             *
*                                                                               *
* The `urlpatterns` list routes URLs to views. For more information please see: *
*    https://docs.djangoproject.com/en/1.11/topics/http/urls/                   *
* Examples:                                                                     *
* Function views                                                                *
*   1. Add an import:  from my_app import views                                 *
*   2. Add a URL to urlpatterns:  url(r'^$', views.home, name='home')           *
* Class-based views                                                             *
*   1. Add an import:  from other_app.views import Home                         *
*   2. Add a URL to urlpatterns:  url(r'^$', Home.as_view(), name='home')       *
* Including another URLconf                                                     *
*   1. Import the include() function: from django.conf.urls import url, include *
*   2. Add a URL to urlpatterns:  url(r'^blog/', include('blog.urls'))          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.urls import include, path

from black_widow.app.gui.web.wsgi import WEB_PACKAGE


urlpatterns = [
    path('', include(WEB_PACKAGE + '.black_widow.urls'))
]



"""
*********************************************************************************
*                                                                               *
* storage.py -- Useful methods and classes to manage files and directory.       *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os
import re
import shutil
import time

from .util import replace_regex


# @return true se il file contiene la stringa find, False altrimenti
def file_contains(find, file):
    # if APP_DEBUG: Log.info('CALLED: file_contains(' + find + ', ' + file + ')')
    if not os.path.isfile(file):
        return False
    with open(file) as f:
        s = f.read()
        return find in s


# @return string il contenuto del file
def read_file(file):
    # if APP_DEBUG: Log.info('CALLED: read_file(' + file + ')')
    if not os.path.isfile(file):
        return ""
    with open(file) as f:
        content = f.read()
        if type(content) == bytes:
            # noinspection PyUnresolvedReferences
            content = str(content.decode('utf-8'))
        return content.rstrip('\n')


# @return True se il file contiene l'espressione regolare regex, False altrimenti
def file_contains_regex(regex, file):
    # if APP_DEBUG: Log.info('CALLED: file_contains_regex(' + regex + ', ' + file + ')')
    if not os.path.isfile(file):
        return False
    reg = re.compile(regex)
    with open(file, 'r') as f:
        text = f.read()
    matches = re.findall(reg, text)
    return len(matches) > 0


def replace_in_file(find, replace, file):
    """
    Esegue il replace della stringa che trova, con la stringa nel parametro replace
    :param find: la stringa da trovare
    :param replace: la stringa che andra' a sostituire la stringa trovata
    :param file: il file in cui sovrascrivere find con replace
    :return: True se trova una stringa find, False altrimenti
    """
    # if APP_DEBUG:
    #    Log.info('CALLED: replace_in_file(' + find + ', ' + replace + ', ' + file + ')')
    if find == replace:
        return False
    if not file_contains(find, file):
        return False
    # Safely write the changed content, if found in the file
    with open(file, 'r') as f:
        t = f.read()
    with open(file, 'w') as f:
        s = t.replace(find, replace)
        f.write(s)
    return True


# Esegue il replace della regex che trova, con la stringa replace
# @param regex l'espressione regolare da trovare
# @param replace la stringa che andra' a sostituire la regex trovata
# @param file il file in cui sovrascrivere regex con replace
# @return True se trova una regex equivalente ad una stringa diversa da
#         replace, False altrimenti
def replace_in_file_regex(regex, replace, file):
    # if APP_DEBUG:
    #    Log.info('CALLED: replace_in_file_regex(' + regex + ', ' + replace + ', ' + file + ')')
    if not os.path.isfile(file):
        with open(file, 'a') as f:
            f.close()
    with open(file, 'r') as f:
        content = f.read()
    content_new = replace_regex(regex, replace, content)  # re.sub(regex, replace, content, flags = re.M)
    if content != content_new:
        overwrite_file(content_new, file)
        return True
    return False


# Sovrascrive il contenuto del file con content
def overwrite_file(content, file):
    # if APP_DEBUG:
    #    Log.info('CALLED: overwrite_file(' + content + ', ' + file + ')')
    with open(file, 'w') as f:
        f.write(str(content) + '\n')


# Appende content nel file
def append_in_file(content, file):
    with open(file, 'a') as f:
        f.write(str(content) + '\n')


# Se la cartella folder non esiste, la crea
def check_folder(folder):
    if not os.path.isdir(folder):
        os.makedirs(folder)


# Elimina tutti i files presenti nella cartella passata per argomento
def clean_folder(folder):
    # if APP_DEBUG:
    #    Log.info('CALLED: clean_folder(' + folder + ')')
    if os.path.isdir(folder):
        for file in os.listdir(folder):
            file_path = os.path.join(folder, file)
            if file_path != folder:
                delete(file_path)
        return True
    return False


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


# Copia il file o la cartella "src", in "dest"
def copy(src, dest):
    # if APP_DEBUG:
    #    Log.info('CALLED: copy(' + src + ', ' + dest + ')')
    dest_parent = os.path.dirname(dest)
    check_folder(dest_parent)
    if os.path.exists(dest):
        delete(dest)
    if os.path.isdir(src):
        shutil.copytree(src, dest)
    else:
        shutil.copy2(src, dest)


# Muove il file o la cartella "src", in "dest"
def move(src, dest):
    # if (APP_DEBUG): Log.info('CALLED: move(' + src + ', ' + dest + ')')
    dest_parent = os.path.dirname(dest)
    check_folder(dest_parent)
    if os.path.exists(dest):
        delete(dest)
    shutil.move(src, dest)


def ls(directory: str) -> list:
    """
    Returns a list containing the names of the files in the directory.
    :param directory: The directory
    :return: The list of the files in the directory
    """
    return [os.path.join(directory, entry) for entry in os.listdir(directory)]


def basename(path: str) -> str:
    """
    Returns the final component of a pathname.
    :param path: The full path
    :return: the final component of a pathname
    """
    return os.path.basename(path)


def is_file(path: str) -> bool:
    """
    Test whether a path is a regular file.
    :param path: The file path
    :return: True if path is a file, otherwise False
    """
    return os.path.isfile(path)


# Elimina il file o la cartella passato per argomento
# noinspection PyBroadException
def delete(file):
    if file is None:
        return True
    attempts = 0
    while os.path.exists(file):
        try:
            if os.path.isfile(file):
                os.remove(file)
            elif os.path.islink(file):
                os.unlink(file)
            elif os.path.isdir(file):
                shutil.rmtree(file)
        except PermissionError:
            return False
        except (Exception, IOError):
            attempts += 1
            if attempts >= 5:
                return False
            time.sleep(0.5)
    return True


def chmod(path, mode, recursive=False):
    os.chmod(path, mode)
    if not recursive:
        return
    for root, dirs, files in os.walk(path):
        for d in dirs:
            os.chmod(os.path.join(root, d), mode)
        for f in files:
            os.chmod(os.path.join(root, f), mode)



"""
*********************************************************************************
*                                                                               *
* wgsi.py -- WSGI config for web project.                                       *
*                                                                               *
* It exposes the WSGI callable as a module-level variable named ``application`` *
*                                                                               *
* For more information on this file, see                                        *
* https://docs.djangoproject.com/en/1.11/howto/deployment/wsgi/                 *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os
import sys

from os.path import dirname
from django.core.wsgi import get_wsgi_application


WEB_PACKAGE = 'black_widow.app.gui.web'

root_path = dirname(dirname(dirname(dirname(dirname(__file__)))))    # to access at "black_widow" package
sys.path.insert(0, root_path)

os.environ.setdefault("DJANGO_SETTINGS_MODULE", WEB_PACKAGE + ".settings")

application = get_wsgi_application()



#!/usr/bin/env python3
# coding=utf-8

"""
****************************   🕷️  Black-Widow  🕷️   *****************************
*                                                                               *
* black_widow.py -- Main black-widow file.                                      *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from . import app

import os
import sys


class AppType:
    CMD = 'CMD'
    GUI = 'GUI'


def make_temp_dir():
    app.helpers.storage.check_folder(app.env.APP_TMP)
    try:
        app.helpers.storage.chmod(app.env.APP_TMP, 0o0777, True)
    except PermissionError:
        pass


# Startup
def init(app_type):
    make_temp_dir()
    app.services.Log.info(app.env.APP_NAME + ' ' + str(app_type) + ' started, PID=' + str(os.getpid()))
    app.services.Log.info('DEBUG is ' + str(app.env.APP_DEBUG))
    app.services.Log.info('ROOT_PATH is ' + str(app.env.ROOT_PATH))


# Main function for TEST app
def main_test():
    init(AppType.CMD)
    ip = app.helpers.network.get_ip_address()
    print('ip: ' + str(ip))


# Main function for GUI app
def main_gui():
    init(AppType.GUI)
    app.gui.django_gui()


# Main function for command line app
def main_cmd(arguments):
    init(AppType.CMD)
    if arguments.django:
        django_args = str.split(arguments.django)
        app.services.Log.info('Django manager executed')
        app.services.Log.info('Django arguments: ' + str(django_args))
        app.gui.django_cmd(django_args)
        sys.exit(0)

    elif arguments.pcap:
        if arguments.pcap_int is None:
            print("\nSpecify at least one interface (eg. --pcap-int=wlan0)\n")
            sys.exit(1)
        app.managers.sniffer.PcapSniffer.sniff(src_file=arguments.pcap_src, interfaces=arguments.pcap_int,
                                               dest_file=arguments.pcap_dest, filters=arguments.pcap_filters,
                                               limit_length=arguments.pcap_limit, pkt_count=arguments.pcap_count,
                                               callback=None)
    elif arguments.sql:
        if arguments.sql_url is None:
            print('Specify an url! (eg. --sql-url=https://black-widow.io)\n')
            sys.exit(1)
        if not app.helpers.validators.is_url(arguments.sql_url):
            print('"' + arguments.sql_url + '" is not a valid URL!\n')
            sys.exit(1)
        app.managers.injection.SqlInjection.inject(
            arguments.sql_forms,
            arguments.sql_url,
            arguments.sql_depth
        )
    elif arguments.crawl:
        if arguments.crawl_url is None:
            print('Specify an url! (eg. --crawl-url=https://black-widow.io)\n')
            sys.exit(1)
        if not app.helpers.validators.is_url(arguments.crawl_url):
            print('"' + arguments.crawl_url + '" is not a valid URL!\n')
            sys.exit(1)
        crawl_types = app.managers.parser.HtmlParser.types()
        if arguments.crawl_type not in crawl_types:
            print('Specify a crawling type! (eg. --crawl-type=form_parse)\n')
            print('Accepted types:')
            type_desc = app.managers.parser.HtmlParser.type_descriptions()
            for crawl_type in crawl_types:
                print('    - ' + crawl_type + ' (' + type_desc[crawl_type] + ')')
            print('')
            sys.exit(1)
        app.managers.parser.HtmlParser.crawl(
            arguments.crawl_url,
            arguments.crawl_type,
            app.managers.parser.HtmlParser.print_parsed,
            arguments.crawl_depth
        )


# Main function generic app
def main():
    arguments = app.arguments.get_arguments()
    if arguments.gui:
        main_gui()
    else:
        main_cmd(arguments)



"""
*********************************************************************************
*                                                                               *
* abstract_view.py -- Django Abstract view.                                     *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import itertools
import math

from os.path import join
from django.views.generic import TemplateView

from black_widow.app.helpers.util import is_listable, timestamp
from black_widow.app.helpers import storage
from black_widow.app.env import APP_TMP


class AbstractView(TemplateView):
    """
    Abstract view extended by application views
    """
    name = None
    error_templates = {
        'root_required': 'error/root_required.html'
    }

    def upload_file(self, tmp_file) -> str:
        """
        :param tmp_file: the in-memory uploaded file
        :type tmp_file: django.core.files.uploadedfile.InMemoryUploadedFile
        :rtype: str
        """
        upload_folder = join(APP_TMP, self.name)
        storage.check_folder(upload_folder)
        uploaded_filename = timestamp() + '_' + tmp_file.name
        uploaded_path = join(upload_folder, uploaded_filename)
        storage.delete(uploaded_path)
        with open(uploaded_path, 'wb+') as destination:
            for chunk in tmp_file.chunks():
                destination.write(chunk)
        return str(uploaded_path)

    def session_put(self, session, value: dict):
        """
        :type session: django.contrib.sessions.backends.db.SessionStore
        :type value: dict
        """
        session[self.name] = value

    def session_update(self, session, value: dict):
        """
        :type session: django.contrib.sessions.backends.db.SessionStore
        :type value: dict
        """
        session_value = session.get(self.name)
        if type(session_value) is not dict:
            session_value = value
        else:
            session_value.update(value)
        session[self.name] = session_value

    def session_get(self, session, params=None):
        """
        :type session: django.contrib.sessions.backends.db.SessionStore
        :type params: dict
        :rtype: dict
        """
        if params is None:
            params = {}
        session_params = session.get(self.name)
        if session_params is None:
            session_params = dict()
        return_params = session_params
        for key, values in params.items():
            session_value = session_params.get(key)
            if is_listable(values):
                return_values = dict()
                for value in values:
                    return_values[value] = (value == session_value)
            else:
                return_values = {
                    {
                        values: (session_value is not None)
                    }
                }
            return_params[key] = return_values
        return return_params

    @staticmethod
    def pagination(elements: dict, page: int, page_size: int):
        try:
            page = int(page)
        except TypeError:
            page = 1
        try:
            page_size = int(page_size)
        except TypeError:
            page_size = 10
        elements_tot = len(elements)
        start = page_size * (page - 1)
        stop = start + page_size
        page_end = math.ceil(elements_tot / page_size)
        result = dict(itertools.islice(elements.items(), start, stop))
        return {
            'result': result,
            'page': page,
            'page_start': 1,
            'page_end': page_end,
            'page_size': page_size,
            'total': elements_tot
        }



# -*- coding: utf-8 -*-

"""
*********************************************************************************
*                                                                               *
* validators.py -- Validation methods.                                          *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import re

from black_widow.app.helpers.util import regex_is_string


def is_ip(ip_address: str) -> bool:
    """
    Checks if the input ip_address is a valid ip address
    :param ip_address: The ip address to check
    :return: True, if text is valid ip address, otherwise False
    """
    if type(ip_address) != str:
        return False
    regex = '^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$'
    reg = re.compile(regex)
    return reg.match(ip_address) is not None


def is_mac(mac_address: str) -> bool:
    """
    Checks if the input mac_address is a valid mac address
    :param mac_address: The mac address to check
    :return: True, if text is valid mac address, otherwise False
    """
    if type(mac_address) != str:
        return False
    regex = '^\w{2}:\w{2}:\w{2}:\w{2}:\w{2}:\w{2}$'
    reg = re.compile(regex)
    return reg.match(mac_address) is not None


def is_url(url: str) -> bool:
    """
    Checks if the input text is an url
    :param url: The string to check
    :return: True, if the string is an url, otherwise False
    """
    if type(url) != str:
        return False
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)
    if regex.match(url):
        return True
    return False


def is_hex(text: str) -> bool:
    """
    Checks if the input text is an hexadecimal value
    :param text: The text to check
    :return: True, if text is an hexadecimal value, otherwise False
    """
    return regex_is_string('^0x[A-Fa-f0-9]+$', text)


def is_int(text: str) -> bool:
    """
    Checks if the input text is a decimal value
    :param text: The text to check
    :return: True, if text is an decimal value, otherwise False
    """
    return regex_is_string('^[0-9]+$', text)



"""
*********************************************************************************
*                                                                               *
* abstract_web_parsing_view.py -- Django Abstract Web Parsing view              *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import os

from black_widow.app.env import APP_STORAGE_OUT
from black_widow.app.gui.web.black_widow.models import WebParsingJobModel
from black_widow.app.helpers.storage import check_folder
from black_widow.app.helpers.util import now
from black_widow.app.managers.parser import HtmlParser
from black_widow.app.services import JsonSerializer, MultiTask

from black_widow.app.gui.web.black_widow.views.abstract_job_view import AbstractJobView


class AbstractWebParsingView(AbstractJobView):
    """
    Abstract Web Parsing View
    """
    storage_out_dir = os.path.join(APP_STORAGE_OUT, 'web_parsing')
    check_folder(storage_out_dir)
    model_class = WebParsingJobModel

    if not os.access(storage_out_dir, os.X_OK):
        os.chmod(storage_out_dir, 0o0755)

    def _new_job(
            self,
            url: str,
            parsing_type: str,
            depth: int or str,
            tags: str,
            cookies: str
    ) -> WebParsingJobModel:
        if parsing_type == WebParsingJobModel.TYPE_SINGLE_PAGE:
            depth = 0
        else:
            depth = int(depth)

        job = WebParsingJobModel()
        job.url = url
        job.parsing_type = parsing_type
        job.parsing_tags = tags
        job.depth = depth
        job.json_file = os.path.join(self.storage_out_dir, now() + '_WEB_PARSING_.json')
        job.cookies = cookies

        def _web_parsing_callback(parsed_page: dict):
            """
            The callback function of crawler.
            This method writes the parsed pages in a json file
            :param parsed_page: The parsed page
            """
            JsonSerializer.add_item_to_dict(parsed_page.get('url'), parsed_page, job.json_file)

        job.pid_file = MultiTask.multiprocess(
            HtmlParser.crawl,
            (
                url,
                tags,
                _web_parsing_callback,
                depth,
                cookies,
            ),
            asynchronous=True,
            cpu=1
        )

        job.save()
        return job

    def _copy_job(self, job: WebParsingJobModel) -> WebParsingJobModel:
        return self._new_job(
            job.url,
            job.parsing_type_key(),
            job.depth,
            job.parsing_tags_key(),
            job.cookies
        )



"""
*********************************************************************************
*                                                                               *
* urls.py -- Django urls of black-widow.                                        *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.urls import path, re_path

from black_widow.app.gui.web.settings import STATIC_URL
from . import views

urlpatterns = [
    path('', views.index, name='black widow'),

    # --- SNIFFING --- #
    path('sniffing', views.Sniffing.SettingsView.as_view(), name='sniffing'),
    path('sniffing/capture', views.Sniffing.CaptureView.as_view(), name='sniffing'),

    # --- WEB PARSING --- #
    path('web/parsing', views.WebParsing.SettingsView.as_view(), name='web parsing'),
    path('web/parsing/parse', views.WebParsing.ParseView.as_view(), name='web parsing'),
    path('web/parsing/download', views.WebParsing.DownloadView.as_view(), name='web parsing'),

    # --- SQL INJECTION --- #
    path('sql', views.Sql.SettingsView.as_view(), name='sql injection'),
    path('sql/inject', views.Sql.InjectView.as_view(), name='sql injection'),

    path('user', views.user, name='user'),
    path('tables', views.tables, name='tables'),
    path('typography', views.typography, name='typography'),
    path('icons', views.icons, name='icons'),
    path('notifications', views.notifications, name='notifications'),
    path('upgrade', views.upgrade, name='upgrade'),

    # static (non debug compatibility without web server)
    re_path(r'^' + STATIC_URL[1:] + '(?P<path>.*)$', views.static, name='static'),
]



"""
*********************************************************************************
*                                                                               *
* web_parsing_view.py -- Django Web Parsing view.                               *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from django.http import JsonResponse, FileResponse
from django.shortcuts import render, redirect

from black_widow.app.gui.web.black_widow.models import WebParsingJobModel
from black_widow.app.gui.web.black_widow.views.web.parsing.abstract_web_parsing_view import AbstractWebParsingView
from black_widow.app.services import Log


class WebParsing:
    """
    Web Parsing Container View
    """
    class SettingsView(AbstractWebParsingView):
        """
        Web Parsing Settings View
        """
        name = 'web parsing'
        template_name = 'web/parsing/settings.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """
            view_params = {
                'types': WebParsingJobModel.TYPES,
                'tags': WebParsingJobModel.PARSE_TAGS,
                'jobs': WebParsingJobModel.all()
            }
            return render(request, self.template_name, view_params)

        def post(self, request):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponseRedirect
            """
            request_params: dict = request.POST.dict()
            job = self._new_job(
                request_params.get('url'),
                request_params.get('type'),
                request_params.get('depth'),
                request_params.get('tags'),
                request_params.get('cookies')
            )
            return redirect('/web/parsing/parse?id=' + str(job.id))

    class ParseView(AbstractWebParsingView):
        """
        Web Parsing Parse View
        """
        name = 'web parsing'
        template_name = 'web/parsing/parse.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """
            return self._get_job(request, redirect_url='/web/parsing')

        def post(self, request) -> JsonResponse:
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.JsonResponse
            """
            return self._post_job(request)

    class DownloadView(AbstractWebParsingView):
        """
        Web Parsing Parse View
        """
        name = 'web parsing'
        template_name = 'web/parsing/parse.html'

        def get(self, request, *args, **kwargs):
            """
            :type request: django.core.handlers.wsgi.WSGIRequest
            :return: django.http.HttpResponse
            """

            request_params: dict = request.GET.dict()

            try:
                job_id = int(request_params.get('id'))
            except (ValueError, TypeError) as e:
                Log.error(str(e))
                return redirect('/web/parsing')

            job = self._get_job_model(job_id)
            if job is None:
                return redirect('/web/parsing/parse?id=' + str(job_id))

            return FileResponse(open(job.json_file, 'rb'), as_attachment=True)



"""
*********************************************************************************
*                                                                               *
* sqlmapcli.py -- sqlmap client.                                                *
*                                                                               *
* Class to interface with the sqlmap local server.                              *
*                                                                               *
* sqlmap repository:                                                            *
* https://github.com/sqlmapproject/sqlmap                                       *
*                                                                               *
* sqlmap license:                                                               *
* https://raw.githubusercontent.com/sqlmapproject/sqlmap/master/LICENSE         *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from sqlmap.lib.utils.api import server as sqlmap_server
from time import sleep

from black_widow.app.managers.request import HttpRequest
from black_widow.app.services import Log, MultiTask
from black_widow.app.helpers.network import check_socket
from black_widow.app.helpers.util import rand_str

from .sqlmaptask import SqlmapTask


class SqlmapClient:
    """
    Sqlmap Client
    """

    DEFAULT_HOST = '127.0.0.1'
    DEFAULT_PORT = 8775

    _client = None

    def __init__(self, host: str = DEFAULT_HOST, port: int = DEFAULT_PORT):
        """
        :param host: The host
        :param port: The port
        """
        self.host = host
        self.port = port
        self.base_url = 'http://' + self.host + ':' + str(port)
        # Start the sqlmap-api server in a parallel thread
        Log.info("Starting sqlmap-api server in a parallel thread")
        MultiTask.multithread(sqlmap_server, (self.host, self.port), True, 1)
        while not check_socket(self.host, self.port):
            # Wait sqlmap-api server
            sleep(0.1)
        Log.success("Sqlmap-api server started!")

    """ Public static methods """

    @staticmethod
    def task_list() -> dict:
        """
        :return: The dictionary of existent tasks
        """
        client = SqlmapClient._get_client()
        return SqlmapTask.task_list(client.base_url)

    @staticmethod
    def task_flush():
        """
        Flush task spool (delete all tasks)
        """
        client = SqlmapClient._get_client()
        return SqlmapTask.task_flush(client.base_url)

    @staticmethod
    def try_inject_url(
            url: str,
            cookies: str = '',
            delay: int = 1,
            random_agent: bool = False
    ) -> SqlmapTask:
        task_options = {
            'dbms': 'MySQL',
            'cookie': cookies,
            'referer': url,
            'delay': delay,
            'randomAgent': random_agent,
            'url': url,
        }
        sqlmap_task = SqlmapClient._task_new(url)
        sqlmap_task.option_set(task_options)
        sqlmap_task.scan_start()
        return sqlmap_task

    @staticmethod
    def try_inject_form(
            url: str,
            form: dict,
            cookies: str = '',
            delay: int = 1,
            random_agent: bool = False
    ) -> SqlmapTask:
        # noinspection PyUnresolvedReferences
        return list(SqlmapClient.try_inject_forms(
            {
                url: form
            },
            cookies,
            delay,
            random_agent
        ).values())[0]

    @staticmethod
    def try_inject_forms(
            forms: dict,
            cookies: str = '',
            delay: int = 1,
            random_agent: bool = False
    ) -> dict:
        """
        Try injection with all provided forms
        :param forms: dict A dictionary of { "<url>": [ <parsed_form_1>, <parsed_form_2>, ... ], ... }
        :param cookies: str the request cookies
        :param delay: int The delay on each request
        :param random_agent: True if set a random agent for each sqlmap request
        :rtype: dict
        """
        sqlmap_tasks = dict()
        Log.info('Trying injection with cookies: ' + str(cookies))
        for url, page_forms in forms.items():
            page_forms: dict    # The forms in page returned by url
            if page_forms.get('tag') == 'form':
                page_forms_in_loop = {
                    0: page_forms
                }
            else:
                page_forms_in_loop = page_forms
            # noinspection PyUnusedLocal
            for page_form in page_forms_in_loop.values():
                page_form: dict    # The attributes and inputs of form
                page_form_attrs = page_form.get('attrs')
                action: str = page_form_attrs.get('action')
                method: str = page_form_attrs.get('method')
                inputs: dict = page_form.get('children')

                if random_agent:
                    agent = None
                else:
                    agent = HttpRequest.default_agent()

                task_options = {
                    'dbms': 'MySQL',
                    'cookie': cookies,
                    'agent': agent,
                    'referer': url,
                    'delay': delay,
                    'randomAgent': random_agent,
                    'method': method,
                    'url': action,
                    'data': SqlmapClient.__get_data(inputs)
                }

                csrf_token = SqlmapClient.__get_csrf_token(inputs)
                if csrf_token is not None:
                    csrf_token_name = csrf_token.get('name')
                    task_options.update({
                        'csrfUrl': url,
                        'csrfMethod': HttpRequest.Type.GET,
                        'csrfToken': csrf_token_name,
                    })

                # for key, value in task_options.items():
                #     print('---------- ' + key + ': ----------')
                #     print(value)

                sqlmap_task = SqlmapClient._task_new(action)
                sqlmap_task.option_set(task_options)
                sqlmap_tasks[sqlmap_task.id] = sqlmap_task

                sqlmap_task.scan_start()

        return sqlmap_tasks

    """ Protected static methods """

    @staticmethod
    def _get_client():
        """
        :rtype: SqlmapClient
        """
        if SqlmapClient._client is None:
            SqlmapClient._client = SqlmapClient()
        return SqlmapClient._client

    @staticmethod
    def _task_new(target_url: str) -> SqlmapTask:
        """
        :param target_url: The target URL
        :return: The new task
        """
        client = SqlmapClient._get_client()
        return SqlmapTask.task_new(client.base_url, target_url)

    @staticmethod
    def __get_csrf_token(inputs: dict) -> dict:
        csrf_field = None
        for name, input_field in inputs.items():
            if input_field.get('type') != 'hidden':
                continue
            name_lower: str = name.lower()
            if 'csrf' in name_lower:
                return input_field
            if 'token' in name_lower:
                csrf_field = input_field
        return csrf_field

    @staticmethod
    def __get_data(inputs: dict) -> str:
        data = ''
        for input_field in inputs.values():
            input_attrs = input_field.get('attrs')
            input_name: str = input_attrs.get('name')
            if input_name is None:
                continue
            input_type: str = input_attrs.get('type')
            input_value: str = input_attrs.get('value')
            if input_type is not None:
                input_type = input_type.lower()
            if input_value is None or input_value == '':
                if input_type == 'email':
                    input_value = 'email@example.com'
                elif type == 'password':
                    input_value = rand_str()
                else:
                    input_value = '1'
            data += input_name + '=' + input_value + '&'
        if len(data) > 0:
            # Remove last "&"
            data = data[0:-1]
        return data



"""
*********************************************************************************
*                                                                               *
* env.py -- Environment variables.                                              *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

from os.path import dirname, join, isfile
from pathlib import Path

# ------- Black-widow files ------- #
APP_PATH = dirname(__file__)    # /path/to/app
ROOT_PATH = dirname(APP_PATH)   # parent of /path/to/app
PRIVATE_ENV_FILE = join(ROOT_PATH, '.env')

# ----- Editable environments ----- #
APP_DEBUG = False
APP_WEB_HOST = '127.0.0.1'
APP_WEB_PORT = 8295
APP_TMP = '/tmp/black-widow'
IGNORE_NON_ROOT = False
FLAG_REGEX = '[A-Z0-9]{31}='

EDITABLE_ENV = (
    'APP_DEBUG',
    'APP_WEB_HOST',
    'APP_WEB_PORT',
    'APP_TMP',
    'IGNORE_NON_ROOT',
    'FLAG_REGEX'
)

# ----------- Read .env ----------- #
if isfile(PRIVATE_ENV_FILE):
    with open(PRIVATE_ENV_FILE) as f:
        for e in f.readlines():
            env = e.strip()
            if len(env) < 3:
                continue
            if env[0] == '#':
                continue
            key, val = env.split('=', 1)[0:2]
            key: str = key.strip()
            val: str = val.strip()
            if len(key) == 0 or len(val) == 0:
                raise EnvironmentError("Wrong environment in .env file: " + env)
            if key not in EDITABLE_ENV:
                raise EnvironmentError("Non editable environment in .env file: " + env)
            if val[0] == "'" or val[0] == '"':
                i = 0
                val_ok = ''
                for char in val:
                    val_ok += val[i]
                    if i >= 1 and char == val[0]:
                        if val[i-1] != '\\':
                            break
                    i += 1
                val = val_ok
                if len(val) == 0:
                    raise EnvironmentError("Wrong environment in .env file: " + env)
            else:
                val = val.split('#')[0].strip()
            exec(key + ' = ' + str(val))

# ----- Derived environments ------ #
# App info
APP_VERSION = '1.7.2#beta'
APP_NAME = 'Black Widow'
APP_PROC = 'black-widow'
# App files
APP_MAIN_FILENAME = APP_PROC + '.py'
APP_LOGFILE = join(APP_TMP, APP_PROC + '.log')
# APP_STORAGE = join(APP_PATH, 'storage')
APP_STORAGE = join(str(Path.home()), '.black_widow')
APP_STORAGE_OUT = join(APP_STORAGE, 'out')
APP_WEB_ROOT = join(APP_PATH, 'gui', 'web')
APP_WEB = join(APP_WEB_ROOT, 'black_widow')
APP_SETTINGS = join(APP_STORAGE, 'settings.json')
EXEC_PATH = join(ROOT_PATH, APP_MAIN_FILENAME)
RES_PATH = join(ROOT_PATH, 'resources')



# -*- coding: utf-8 -*-

"""
*********************************************************************************
*                                                                               *
* pcap_sniffer.py -- Packet Capture (pcap)                                      *
*                                                                               *
* Methods and Classes to sniff the network traffic through pyshark.             *
*                                                                               *
* pyshark repository:                                                           *
* https://github.com/KimiNewt/pyshark                                           *
*                                                                               *
* pyshark license:                                                              *
* https://raw.githubusercontent.com/KimiNewt/pyshark/master/LICENSE.txt         *
*                                                                               *
********************** IMPORTANT BLACK-WIDOW LICENSE TERMS **********************
*                                                                               *
* This file is part of black-widow.                                             *
*                                                                               *
* black-widow is free software: you can redistribute it and/or modify           *
* it under the terms of the GNU General Public License as published by          *
* the Free Software Foundation, either version 3 of the License, or             *
* (at your option) any later version.                                           *
*                                                                               *
* black-widow is distributed in the hope that it will be useful,                *
* but WITHOUT ANY WARRANTY; without even the implied warranty of                *
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                 *
* GNU General Public License for more details.                                  *
*                                                                               *
* You should have received a copy of the GNU General Public License             *
* along with black-widow.  If not, see <http://www.gnu.org/licenses/>.          *
*                                                                               *
*********************************************************************************
"""

import pyshark
import socket

from pyshark.packet.fields import LayerField
from pyshark.packet.layer import Layer
from pyshark.packet.packet import Packet

from black_widow.app.services import Log
from black_widow.app.helpers.util import regex_is_string, root_required, is_executable as exec_is_executable, os_exec
from black_widow.app.helpers.validators import is_hex, is_mac, is_int, is_ip

from .pcap_sniffer_util import MacManufacturer
from .pcap_sniffer_util import PcapLayerField


class PcapSniffer:
    """
    Packet Capture Manager
    """

    _IGNORED_DOMAINS = (MacManufacturer.MANUFACTURERS_DOMAIN,)

    def __init__(
            self,
            filters: str = None,
            src_file: str = None,
            dest_file: str = None,
            interfaces: list = None,
            limit_length: int = None,
            pkt_count: int = None,
            callback=None
    ):
        """
        Packet capture method
        :param filters: https://wiki.wireshark.org/DisplayFilters
        :param src_file: Il file .pcap da cui leggere i pacchetti ascoltati (o None, per Live sniffing)
        :param dest_file: Il file in cui scrivere il .pcap dei pacchetti ascoltati (o None)
        :param interfaces: The list of interfaces to sniff (or None, to sniff all interfaces)
        :param limit_length: The limit length of each packet field (they will be truncated), or None
        :param pkt_count: Max packets to sniff, or None
        :param callback: The callback method to call (or None) (@see PcapSniffer._user_callback_example)
        """
        if not PcapSniffer.is_executable():
            raise RuntimeError('Unable to execute pcap sniffer')
        self.count = 0  # Sniffed packets
        self.max_count = pkt_count
        # Prevents the mac manufacturer lookup sniffing
        self.filters = PcapSniffer._get_filters(filters)
        self.src_file = src_file
        self.dest_file = dest_file
        self.limit_length = limit_length
        self.user_callback = callback
        self.interfaces = interfaces
        Log.info('Analyzing filters: ' + str(self.filters))
        if self.src_file is not None:
            Log.info('Analyzing file: ' + self.src_file)
            self._capture = pyshark.FileCapture(
                input_file=self.src_file,
                display_filter=self.filters,
                output_file=self.dest_file,
                # include_raw=True,
                # use_json=True
                # debug=APP_DEBUG
            )
        else:
            Log.info('Analyzing live traffic')
            self._capture = pyshark.LiveCapture(
                interface=self.interfaces,
                display_filter=self.filters,
                output_file=self.dest_file,
                # include_raw=True,
                # use_json=True
                # debug=APP_DEBUG
            )

    @staticmethod
    def sniff(
            filters: str = None,
            src_file: str = None,
            dest_file: str = None,
            interfaces: list = None,
            limit_length: int = None,
            pkt_count: int = None,
            callback=None
    ):
        """
        Packet capture method
        :param filters: https://wiki.wireshark.org/DisplayFilters
        :param src_file: Il file .pcap da cui leggere i pacchetti ascoltati (o None, per Live sniffing)
        :param dest_file: Il file in cui scrivere il .pcap dei pacchetti ascoltati (o None)
        :param interfaces: The list of interfaces to sniff (or None, to sniff all interfaces)
        :param limit_length: The limit length of each packet field (they will be truncated), or None
        :param pkt_count: Max packets to sniff, or None
        :param callback: The callback method to call (or None)
        :rtype: PcapSniffer
        """
        pcap_sniffer = PcapSniffer(filters, src_file, dest_file, interfaces, limit_length, pkt_count, callback)
        pcap_sniffer.start()
        return pcap_sniffer

    def start(self):
        """
        Starts the capture
        """
        self._capture.apply_on_packets(self._callback, packet_count=self.max_count)

    @staticmethod
    def print_pkt(pkt_dict: dict):
        """
        Print the pkt_dict
        :param pkt_dict: The pkt dict created by PcapSniffer._callback(pkt)
        """
        for key, value in pkt_dict.items():
            if key == 'layers':
                for layer_dict in value:
                    PcapSniffer._print_layer(layer_dict)
            elif key == 'frame_info':
                PcapSniffer._print_layer(value)
            else:
                print(str(key) + ': ' + str(value))

    @staticmethod
    def is_executable() -> bool:
        if not exec_is_executable('tshark') or not exec_is_executable('dumpcap'):
            return False
        tshark_out = os_exec('tshark -c 1')
        for out_line in tshark_out:
            if 'permission' in out_line.lower():
                return False
        dumpcap_out = os_exec('dumpcap -c 1')
        for out_line in dumpcap_out:
            if 'permission' in out_line.lower():
                return False
        return True


    def _callback(self, pkt: Packet):
        """
        :param pkt: The pyshark packet
        """
        pkt_dict = {
            'number': str(pkt.number),
            'time': str(pkt.frame_info.time_relative),
            'length': str(pkt.length),
            'captured_length': str(pkt.captured_length),
            'interface_captured': str(pkt.interface_captured),
            'highest_layer': str(pkt.highest_layer),
            'sniff_time': str(pkt.sniff_time),
            'sniff_timestamp': str(pkt.sniff_timestamp),
            'transport_layer': str(pkt.transport_layer),
            'protocol': str(pkt.highest_layer),
            'frame_info': PcapSniffer._get_layer_dict(pkt.frame_info)[0],
            'layers': []
        }
        for layer in pkt.layers:
            layer_dict, src_dest_dict = PcapSniffer._get_layer_dict(layer)
            pkt_dict['source'] = PcapSniffer._merge_addr(
                pkt_dict.get('source'),
                src_dest_dict.get('source')
            )
            pkt_dict['destination'] = PcapSniffer._merge_addr(
                pkt_dict.get('destination'),
                src_dest_dict.get('destination')
            )
            protocol = src_dest_dict.get('protocol')
            if protocol is not None:
                pkt_dict['protocol'] = protocol
            pkt_dict['layers'].append(layer_dict)
        if self.user_callback is not None:
            self.user_callback(pkt_dict)
        else:
            PcapSniffer.print_pkt(pkt_dict)
        self.count += 1

    # noinspection PyUnusedLocal
    @staticmethod
    def _user_callback_example(pkt_dict: dict) -> None:
        """
        This is an user_callback example method.
        It manages the pkt dictionary
        :param pkt_dict: The packet dictionary
        :return: None
        """
        # Manage pkt_dict
        return None

    @staticmethod
    def _merge_addr(host1: dict, host2: dict):
        """
        Merge host1 and host2 by preferring host2
        :param host1: {
                        'mac': <mac_addr>,
                        'mac_manufacturer': tuple,
                        'ip': <ip_addr>,
                        'ip_host': list
                      }
        :param host2: //
        :return: The host1 merged with host2
        """
        if host1 is None:
            return host2
        if host2 is None:
            return host1
        host = host2.copy()
        for key, val in host2.items():
            if val is not None:
                continue
            host[key] = host1.get(key)
        ip = host.get('ip')
        ip_host = host.get('ip_host')
        mac = host.get('mac')
        mac_manufacturer = host.get('mac_manufacturer')
        if ip is not None:
            host['label'] = ip
            host['title'] = ip_host
        else:
            if mac_manufacturer is None:
                host['label'] = mac
            else:
                host['label'] = mac_manufacturer
                host['title'] = mac
        return host

    @staticmethod
    def _field_is_binary(field: PcapLayerField):
        field_label = field.label
        equal_index = field_label.find(' = ')
        binary_key = field_label[0:equal_index]
        return equal_index >= 0 and regex_is_string('^(\.| |0|1)+$', binary_key)

    # noinspection PyProtectedMember
    @staticmethod
    def _get_layer_dict(layer: Layer) -> (dict, dict):
        """
        TODO: Manage Tags
        :param layer: The layer to process
        :return: The dictionary of layer, The dictionary of src and dest ip/mac (looked up)
        """

        field_tree = dict()  # tree dict { name => pkt }

        def update_field_tree_keys(local_field_tree_keys: list) -> (list, dict):
            """
            Insert into the field_tree the input keys recursively
            Eg. local_field_tree_keys = ['a', 'b', 'c']
                field_tree['a']['b']['c'] = dict()
            :param local_field_tree_keys: the dictionary keys used to create the parent dicts of field
            :return: The family (as list) and the node (as dict) of field to insert
            """
            local_field_node = field_tree
            local_field_tree_family = []
            for local_field_tree_key in local_field_tree_keys:
                local_field_tree_child = local_field_node.get(local_field_tree_key)
                if local_field_tree_child is None:
                    local_field_tree_child = dict()
                    local_field_node[local_field_tree_key] = local_field_tree_child
                local_field_tree_family.insert(0, local_field_tree_child)
                local_field_node = local_field_tree_child
            return local_field_tree_family, local_field_node

        pcap_layer_field_root = PcapLayerField(sanitized_name='root')

        # noinspection PyProtectedMember
        def local_get_field_tree(local_field: LayerField) -> PcapLayerField:
            """
            :param local_field: The LayerField to insert in dict
            """
            try:
                parent_poss = (int(local_field.pos), int(local_field.pos) - int(local_field.size))
            except TypeError:
                parent_poss = ()
                local_field.pos = 0
                local_field.size = 0

            if local_field.name is None:
                local_field.name = ''

            field_tree_keys = local_field.name.split('.')
            family, node = update_field_tree_keys(field_tree_keys)

            def find_pcap_layer_field_parent(local_member: dict, only_hex: bool = False) -> PcapLayerField or None:
                """
                If exists, found the big brother of local_field, otherwise, the parent
                :param local_member: The family tree of local_field (as dict)
                :param only_hex: True if the member_parent of local_field should has an hexadecimal value
                :return: If exits, the big brother of local_field, otherwise, the parent or None
                """
                parent = None
                for key, member_parent in local_member.items():
                    member_parent: PcapLayerField or dict
                    if not only_hex and isinstance(member_parent, dict):
                        # Check brothers
                        member_brother = find_pcap_layer_field_parent(member_parent, True)
                        if member_brother is not None:
                            return member_brother  # brother
                    if key in parent_poss and member_parent.name != local_field.name:
                        member_parent: PcapLayerField
                        if not only_hex:
                            parent = member_parent  # parent (but the preferred is brother)
                        elif is_hex(member_parent.value) and \
                                member_parent.pos == int(local_field.pos) and \
                                not PcapSniffer._field_is_binary(member_parent):
                            return member_parent  # brother
                return parent

            pcap_layer_field_parent = None
            for member in family:
                pcap_layer_field_parent = find_pcap_layer_field_parent(member)
                if pcap_layer_field_parent is not None:
                    break

            if pcap_layer_field_parent is None:
                pcap_layer_field_parent = pcap_layer_field_root

            local_field_sanitized_name = layer._sanitize_field_name(local_field.name)
            local_pcap_layer_field = PcapLayerField(
                local_field,
                local_field_sanitized_name,
                pcap_layer_field_parent
            )
            node[int(local_field.pos)] = local_pcap_layer_field  # Update dictionary tree
            return local_pcap_layer_field

        source = {
            'mac': None,
            'mac_manufacturer': None,
            'mac_lookup': None,
            'ip': None,
            'ip_host': None,
            'port': None
        }
        destination = {
            'mac': None,
            'mac_manufacturer': None,
            'mac_lookup': None,
            'ip': None,
            'ip_host': None,
            'port': None
        }
        protocol = None

        field_insert = set()
        for field in layer._get_all_fields_with_alternates():
            field: LayerField
            if field.name in PcapLayerField.AMBIGUOUS_FIELD_NAMES:
                continue
            field_unique_key = str(field.pos) + '_' + str(field.name)
            if field_unique_key in field_insert:
                continue
            pcap_layer_field: PcapLayerField = local_get_field_tree(field)
            if pcap_layer_field is None:
                continue

            if pcap_layer_field.sanitized_name in PcapLayerField.PROTO_FIELDS:
                protocol = pcap_layer_field.value
            else:
                host = None
                if pcap_layer_field.sanitized_name in PcapLayerField.SRC_FIELDS:
                    host = source
                elif pcap_layer_field.sanitized_name in PcapLayerField.DST_FIELDS:
                    host = destination
                if host is not None:
                    if is_mac(pcap_layer_field.value):
                        mac_manufacturer_result = MacManufacturer.lookup(pcap_layer_field.value)
                        host['mac'] = pcap_layer_field.value
                        host['mac_manufacturer'] = mac_manufacturer_result.get('manufacturer')
                        # noinspection PyTypeChecker
                        host['mac_lookup'] = mac_manufacturer_result
                    elif is_ip(pcap_layer_field.value):
                        try:
                            host['ip'] = pcap_layer_field.value
                            # noinspection PyTypeChecker
                            host['ip_host'] = socket.gethostbyaddr(pcap_layer_field.value)[0]
                        except (socket.herror, socket.gaierror):
                            pass
                    elif is_int(pcap_layer_field.value):
                        # It's the port
                        host['port'] = pcap_layer_field.value

            field_insert.add(field_unique_key)

        return {
                   'name': layer.layer_name.upper(),
                   'fields': pcap_layer_field_root.get_dict().get('children')
               }, {
                   'source': source,
                   'destination': destination,
                   'protocol': protocol
               }

    @staticmethod
    def _get_filters(user_filters: str):
        if type(user_filters) is not str:
            user_filters = ''
        ignored_hosts = PcapSniffer._ignored_hosts()
        filters = ''
        for host in ignored_hosts:
            filters += 'ip.src != ' + host + ' and ip.dst != ' + host + ' and '
        if user_filters == '' and len(filters) > 0:
            filters = filters[:-5]  # remove last " and "
        else:
            filters += '(' + user_filters + ')'
        return filters

    @staticmethod
    def _ignored_hosts() -> tuple:
        ignored_hosts = ()
        for domain in PcapSniffer._IGNORED_DOMAINS:
            try:
                ignored_hosts += (socket.gethostbyname(domain),)
            except (socket.herror, socket.gaierror):
                pass
        return ignored_hosts

    @staticmethod
    def _print_layer(layer_dict: dict):
        """
        Print the layer_dict
        :param layer_dict: The layer dict created by PcapSniffer._callback(pkt)
        """
        print('Layer: ' + str(layer_dict.get('name')))
        for field_dict in layer_dict.get('fields'):
            PcapSniffer._print_field(field_dict)

    @staticmethod
    def _print_field(field_dict: dict, depth: int = 0):
        """
        Print the field_dict
        :param field_dict: The field dict returned by PcapLayerField.get_dict()
        """
        field_header = '   |' * depth + '   ├── '
        field_label = field_dict.get('label')
        field_key = field_header + '[ ' + str(field_label) + ' ]'
        field_value = field_dict.get('value')
        if field_value is None:
            field_value = ''
        if len(field_value) > 30:
            field_value = field_value[0:30] + '...'
        else:
            field_key += ' = '
        alternate_values = field_dict.get('alternate_values')
        if alternate_values is not None:
            field_key_len = len(field_key)
            for alternate_value in field_dict.get('alternate_values'):
                field_value += '\n' + (field_key_len * ' ') + alternate_value
        pos = str(field_dict.get('pos'))
        size = str(field_dict.get('size'))
        name = str(field_dict.get('name'))
        print(field_key + field_value + ' (pos=' + pos + ', size=' + size + ', name=' + name + ')')
        for field_child in field_dict.get('children'):
            PcapSniffer._print_field(field_child, depth + 1)



# Generated by Django 4.0.4 on 2022-04-29 18:11

from django.db import migrations, models
import django.utils.timezone
import signal


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='SniffingJobModel',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('json_file', models.CharField(max_length=250)),
                ('status', models.PositiveIntegerField(default=signal.Signals['SIGCONT'])),
                ('pid', models.PositiveIntegerField()),
                ('_pid_file', models.CharField(max_length=250)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('filters', models.TextField(null=True)),
                ('_interfaces', models.TextField()),
                ('pcap_file', models.CharField(max_length=250, null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='WebParsingJobModel',
            fields=[
                ('id', models.BigAutoField(primary_key=True, serialize=False)),
                ('json_file', models.CharField(max_length=250)),
                ('status', models.PositiveIntegerField(default=signal.Signals['SIGCONT'])),
                ('pid', models.PositiveIntegerField()),
                ('_pid_file', models.CharField(max_length=250)),
                ('created_at', models.DateTimeField(default=django.utils.timezone.now, editable=False)),
                ('url', models.CharField(max_length=512)),
                ('_parsing_type', models.CharField(choices=[('single_page', 'Single Page'), ('website_crawling', 'Website Crawling')], max_length=50)),
                ('_parsing_tags', models.CharField(choices=[('all_parse', 'All Tags'), ('relevant_parse', 'Relevant Tags (a, script, link, Form Tags, Meta Tags)'), ('form_parse', 'Form Tags (form, input, textarea, select, option)'), ('meta_parse', 'Meta Tags (meta, xmp)')], max_length=50)),
                ('depth', models.IntegerField()),
                ('cookies', models.TextField(null=True)),
            ],
            options={
                'abstract': False,
            },
        ),
    ]
